# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous functions and decorators."""

import aeidon
import gaupol
import inspect
import sys
import traceback
import webbrowser

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk


def char_to_px(nchar, font=None):
    """Convert characters to pixels."""
    if nchar < 0: return nchar
    label = Gtk.Label(label="etaoin shrdlu")
    gaupol.style.use_font(label, font)
    label.show()
    width = label.get_preferred_width()[1]
    return int(round(nchar * width/len(label.props.label)))

def delay_add(delay, function, *args, **kwargs):
    """Call `function` with `args` and `kwargs` once after `delay` (ms)."""
    def call_function(*args, **kwargs):
        function(*args, **kwargs)
        return False # to not be called again.
    return GLib.timeout_add(delay, call_function, *args, **kwargs)

def document_to_text_field(doc):
    """Return :attr:`gaupol.fields` item corresponding to `doc`."""
    if doc == aeidon.documents.MAIN:
        return gaupol.fields.MAIN_TEXT
    if doc == aeidon.documents.TRAN:
        return gaupol.fields.TRAN_TEXT
    raise ValueError("Invalid document: {!r}"
                     .format(doc))

def flash_dialog(dialog):
    """Run `dialog`, destroy it and return response."""
    response = dialog.run()
    dialog.destroy()
    return response

def get_content_size(widget, font=None):
    """Return the width and height desired by `widget`."""
    if isinstance(widget, Gtk.TextView):
        return get_text_view_size(widget, font)
    if isinstance(widget, Gtk.TreeView):
        return get_tree_view_size(widget, font)
    raise ValueError("Unsupported container type: {!r}"
                     .format(type(widget)))

def get_default_player():
    """Return the default video player to use for preview."""
    players = [aeidon.players.MPV, aeidon.players.MPLAYER, aeidon.players.VLC]
    players = [x for x in players if x.found]
    return players[0] if players else aeidon.players.MPV

def get_font():
    """Return custom font or blank string."""
    return (gaupol.conf.editor.custom_font if
            gaupol.conf.editor.use_custom_font and
            gaupol.conf.editor.custom_font else "")

def get_gspell_version():
    """Return :mod:`Gspell` version number as string or ``None``."""
    try:
        # XXX: The full version number is not available.
        # https://gitlab.gnome.org/GNOME/gspell/issues/8
        from gi.repository import Gspell
        return str(Gspell._version)
    except Exception:
        return None

def get_gst_version():
    """Return :mod:`Gst` version number as string or ``None``."""
    try:
        from gi.repository import Gst
        return ".".join(map(str, Gst.version()))
    except Exception:
        return None

def get_icon_image(name, fallback, size):
    """Return icon image from `name` or `fallback` in theme."""
    theme = Gtk.IconTheme.get_default()
    if theme.has_icon(name):
        return Gtk.Image(icon_name=name, icon_size=size)
    return Gtk.Image(icon_name=fallback, icon_size=size)

def get_preview_command():
    """Return command to use for lauching video player."""
    if gaupol.conf.preview.use_custom_command:
        return gaupol.conf.preview.custom_command
    if gaupol.conf.preview.force_utf_8:
        return gaupol.conf.preview.player.command_utf_8
    return gaupol.conf.preview.player.command

def get_text_view_size(text_view, font=None):
    """Return the width and height desired by `text_view`."""
    text_buffer = text_view.get_buffer()
    start, end = text_buffer.get_bounds()
    text = text_buffer.get_text(start, end, False)
    label = Gtk.Label(label=text)
    gaupol.style.use_font(label, font)
    label.show()
    return (label.get_preferred_width()[1]
            + text_view.get_left_margin()
            + text_view.get_right_margin(),
            label.get_preferred_height()[1])

def get_tree_view_size(tree_view, font=None):
    """Return the width and height desired by `tree_view`."""
    scroller = tree_view.get_parent()
    policy = scroller.get_policy()
    scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
    width = scroller.get_preferred_width()[1]
    height = scroller.get_preferred_height()[1]
    scroller.set_policy(*policy)
    return width, height

@aeidon.deco.once
def get_zebra_color(tree_view):
    """Return background color to use for tree view zebra-stripes."""
    # XXX: Zebra stripes would be faster and cleaner done with CSS
    # selectors :nth-child(odd) and :nth-child(even), but they don't
    # seem to work, might even be deliberately broken.
    # https://bugzilla.gnome.org/show_bug.cgi?id=709617#c1
    style = tree_view.get_style_context()
    fg = style.get_color(Gtk.StateFlags.NORMAL)
    bg = style.get_background_color(Gtk.StateFlags.NORMAL)
    color = Gdk.RGBA()
    color.red   = 0.92 * bg.red   + 0.08 * fg.red
    color.green = 0.92 * bg.green + 0.08 * fg.green
    color.blue  = 0.92 * bg.blue  + 0.08 * fg.blue
    return(color)

@aeidon.deco.once
def gst_available():
    """Return ``True`` if :mod:`Gst` and needed plugins are available."""
    try:
        from gi.repository import Gst
    except Exception:
        return False
    if not Gst.ElementFactory.find("playbin"):
        print("GStreamer found, but playbin missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("textoverlay"):
        print("GStreamer found, but textoverlay missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("timeoverlay"):
        print("GStreamer found, but timeoverlay missing.",
              "Try installing gst-plugins-base.",
              file=sys.stderr)
        return False
    if not Gst.ElementFactory.find("gtksink"):
        print("GStreamer found, but gtksink missing.",
              "Try installing gst-plugins-good.",
              file=sys.stderr)
        return False
    if Gst.ElementFactory.find("vaapisink"):
        print("GStreamer-vaapi found, known to sometimes cause trouble.",
              "If video doesn't play, please try uninstalling it.",
              "See https://github.com/otsaloma/gaupol/issues/79",
              file=sys.stderr)
    return True

def hex_to_rgba(string):
    """Return a :class:`Gdk.RGBA` for hexadecimal `string`."""
    rgba = Gdk.RGBA()
    success = rgba.parse(string)
    if not success:
        raise ValueError("Parsing string {!r} failed".format(string))
    return rgba

def idle_add(function, *args, **kwargs):
    """Call `function` with `args` and `kwargs` when idle."""
    def call_function(*args, **kwargs):
        function(*args, **kwargs)
        return False # to not be called again.
    return GLib.idle_add(call_function, *args, **kwargs)

def install_module(name, obj):
    """
    Install `obj`'s module into the :mod:`gaupol` namespace.

    Typical call is of form::

        gaupol.util.install_module("foo", lambda: None)
    """
    gaupol.__dict__[name] = inspect.getmodule(obj)

def iterate_main():
    """Iterate the GTK main loop while events are pending."""
    while Gtk.events_pending():
        Gtk.main_iteration()

def lines_to_px(nlines, font=None):
    """Convert lines to pixels."""
    if nlines < 0: return nlines
    text = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
    label = Gtk.Label(label=text)
    gaupol.style.use_font(label, font)
    label.show()
    height = label.get_preferred_height()[1]
    return int(round(nlines * height))

def new_hbox(spacing):
    """Return a new horizontal :class:`Gtk.Box`."""
    return Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL,
                   spacing=spacing)

def new_vbox(spacing):
    """Return a new vertical :class:`Gtk.Box`."""
    return Gtk.Box(orientation=Gtk.Orientation.VERTICAL,
                   spacing=spacing)

def pack_start(box, widget, padding=0):
    """Pack widget to box without fill or expand."""
    box.pack_start(widget,
                   expand=False,
                   fill=False,
                   padding=padding)

def pack_start_expand(box, widget, padding=0):
    """Pack widget to box with fill and expand."""
    box.pack_start(widget,
                   expand=True,
                   fill=True,
                   padding=padding)

def pack_start_fill(box, widget, padding=0):
    """Pack widget to box with fill, but no expand."""
    box.pack_start(widget,
                   expand=False,
                   fill=True,
                   padding=padding)

def prepare_text_view(text_view):
    """Set spell-check, line-length margin and font properties."""
    if (gaupol.SpellChecker.available() and
        gaupol.conf.spell_check.inline):
        language = gaupol.conf.spell_check.language
        with aeidon.util.silent(Exception):
            checker = gaupol.SpellChecker(language)
            checker.attach(text_view)
    connect = gaupol.conf.editor.connect
    def update_margin(section, value, text_view):
        if gaupol.conf.editor.show_lengths_edit:
            return gaupol.ruler.connect_text_view(text_view)
        return gaupol.ruler.disconnect_text_view(text_view)
    connect("notify::show_lengths_edit", update_margin, text_view)
    update_margin(None, None, text_view)
    def update_font(section, value, text_view):
        text_view.reset_style()
    gaupol.style.use_font(text_view, "custom")
    connect("notify::use_custom_font", update_font, text_view)
    connect("notify::custom_font", update_font, text_view)
    update_font(None, None, text_view)
    def update_spacing(section, value, text_view):
        if gaupol.conf.editor.show_lengths_cell:
            return text_view.set_pixels_above_lines(2)
        return text_view.set_pixels_above_lines(0)
    connect("notify::show_lengths_cell", update_spacing, text_view)
    update_spacing(None, None, text_view)

def raise_default(expression):
    """Raise :exc:`gaupol.Default` if `expression` evaluates to ``True``."""
    if expression:
        raise gaupol.Default

def rgba_to_hex(color):
    """Return hexadecimal string for :class:`Gdk.RGBA` `color`."""
    return "#{:02x}{:02x}{:02x}".format(
        int(color.red   * 255),
        int(color.green * 255),
        int(color.blue  * 255),
    )

def run_dialog(dialog):
    """Run `dialog` and return response."""
    return dialog.run()

def scale_to_content(widget, min_nchar=0,  max_nchar=32768,
                     min_nlines=0, max_nlines=32768, font=None):
    """Set `widget's` size by content, but limited by `min` and `max`."""
    width, height = get_content_size(widget, font)
    width  = max(width, char_to_px(min_nchar, font))
    width  = min(width, char_to_px(max_nchar, font))
    height = max(height, lines_to_px(min_nlines, font))
    height = min(height, lines_to_px(max_nlines, font))
    parent = widget.get_parent()
    if isinstance(parent, Gtk.ScrolledWindow):
        # Vaguely account for possible scrollbars.
        return parent.set_size_request(width + 24, height + 24)
    widget.set_size_request(width, height)

def scale_to_size(widget, nchar, nlines, font=None):
    """Set `widget`'s size to `nchar` and `nlines`."""
    width  = char_to_px(nchar, font)
    height = lines_to_px(nlines, font)
    parent = widget.get_parent()
    if isinstance(parent, Gtk.ScrolledWindow):
        # Vaguely account for possible scrollbars.
        return parent.set_size_request(width + 24, height + 24)
    widget.set_size_request(width, height)

def separate_combo(store, itr, data=None):
    """Separator function for combo box models."""
    return store.get_value(itr, 0) == gaupol.COMBO_SEPARATOR

def set_cursor_busy(window):
    """Set mouse pointer busy when above window."""
    cursor = window.get_window().get_cursor()
    if (cursor is not None and cursor.get_cursor_type() ==
        Gdk.CursorType.WATCH): return
    cursor = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.WATCH)
    window.get_window().set_cursor(cursor)
    iterate_main()

def set_cursor_normal(window):
    """Set mouse pointer normal when above window."""
    cursor = window.get_window().get_cursor()
    if (cursor is not None and cursor.get_cursor_type() ==
        Gdk.CursorType.LEFT_PTR): return
    cursor = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.LEFT_PTR)
    window.get_window().set_cursor(cursor)
    iterate_main()

def show_exception(exctype, value, tb):
    """A :class:`gaupol.DebugDialog` :attr`sys.excepthook`."""
    traceback.print_exception(exctype, value, tb)
    if not isinstance(value, Exception): return
    try: # to avoid recursion.
        dialog = gaupol.DebugDialog()
        dialog.set_text(exctype, value, tb)
        response = dialog.run()
        dialog.destroy()
        if response == Gtk.ResponseType.NO:
            raise SystemExit(1)
    except Exception:
        traceback.print_exc()

def show_uri(uri):
    """Open `uri` in default application."""
    try:
        return Gtk.show_uri(None, uri, Gdk.CURRENT_TIME)
    except Exception:
        # Gtk.show_uri fails on Windows and some misconfigured installations.
        # GError: No application is registered as handling this file
        # Gtk.show_uri: Operation not supported
        if uri.startswith(("http://", "https://")):
            return webbrowser.open(uri)
        raise # Exception

def text_field_to_document(field):
    """Return :attr:`aeidon.documents` item corresponding to `field`."""
    if field == gaupol.fields.MAIN_TEXT:
        return aeidon.documents.MAIN
    if field == gaupol.fields.TRAN_TEXT:
        return aeidon.documents.TRAN
    raise ValueError("Invalid field: {!r}"
                     .format(field))

def tree_path_to_row(path):
    """
    Convert `path` to a list row integer.

    `path` can be either a :class:`Gtk.Treepath` instance or a string
    representation of it (as commonly used by various callbacks).
    """
    if path is None: return None
    if isinstance(path, Gtk.TreePath):
        return path.get_indices()[0]
    if isinstance(path, str):
        return int(path)
    raise TypeError("Bad type {!r} for path {!r}"
                    .format(type(path), path))

def tree_row_to_path(row):
    """Convert list row integer to a :class:`Gtk.TreePath`."""
    if row is None: return None
    return Gtk.TreePath.new_from_string(str(row))



# -*- coding: utf-8 -*-

# Copyright (C) 2012 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Loading and interacting with video."""

import aeidon
import gaupol
import os
import sys

from aeidon.i18n   import _
from gi.repository import GLib
from gi.repository import Gtk

with aeidon.util.silent(Exception):
    from gi.repository import Gst
    from gi.repository import GstTag



class VideoAgent(aeidon.Delegate):

    """Loading and interacting with video."""

    def __init__(self, master):
        """Initialize an :class:`VideoAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        # Maintain an up-to-date cache of subtitle positions in seconds and
        # subtitle texts in order to allow fast polled updates in video player.
        # This cache must be updated when page or subtitle data changes.
        self._cache = []
        self._update_handlers = []

    def _clear_subtitle_cache(self):
        """Clear subtitle position and text cache."""
        self._cache = []

    def _init_cache_updates(self):
        """Initialize cache updates on application signals."""
        self.connect("page-added",    self._update_subtitle_cache)
        self.connect("page-changed",  self._update_subtitle_cache)
        self.connect("page-closed",   self._update_subtitle_cache)
        self.connect("page-switched", self._update_subtitle_cache)

    def _init_player_toolbar(self):
        """Initialize the video player toolbar."""
        self.player_toolbar = Gtk.Toolbar()
        self.player_toolbar.set_style(Gtk.ToolbarStyle.ICONS)
        if sys.platform == "win32":
            self.player_toolbar.set_icon_size(Gtk.IconSize.MENU)
        # win.play-pause
        button = Gtk.ToolButton(
            label=_("_Play/Pause"), icon_name="media-playback-start")
        button.set_action_name("win.play-pause")
        button.set_tooltip_text(_("Play or pause video"))
        self.player_toolbar.insert(button, -1)
        self.player_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        self.play_button = button
        # win.seek-previous
        button = Gtk.ToolButton(
            label=_("Seek _Previous"), icon_name="media-skip-backward")
        button.set_action_name("win.seek-previous")
        button.set_tooltip_text(_("Seek to the start of the previous subtitle"))
        self.player_toolbar.insert(button, -1)
        # win.seek-next
        button = Gtk.ToolButton(
            label=_("Seek _Next"), icon_name="media-skip-forward")
        button.set_action_name("win.seek-next")
        button.set_tooltip_text(_("Seek to the start of the next subtitle"))
        self.player_toolbar.insert(button, -1)
        self.player_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        # win.seek-backward
        button = Gtk.ToolButton(
            label=_("Seek _Backward"), icon_name="media-seek-backward")
        button.set_action_name("win.seek-backward")
        button.set_tooltip_text(_("Seek backward"))
        self.player_toolbar.insert(button, -1)
        # win.seek-forward
        button = Gtk.ToolButton(
            label=_("Seek _Forward"), icon_name="media-seek-forward")
        button.set_action_name("win.seek-forward")
        button.set_tooltip_text(_("Seek forward"))
        self.player_toolbar.insert(button, -1)
        self.player_toolbar.insert(Gtk.SeparatorToolItem(), -1)
        # Volume button
        self.volume_button = Gtk.VolumeButton()
        self.volume_button.props.use_symbolic = False
        adjustment = self.volume_button.get_adjustment()
        adjustment.set_lower(0)
        adjustment.set_upper(1)
        adjustment.set_value(self.player.volume)
        aeidon.util.connect(self, "volume_button", "value-changed")
        item = Gtk.ToolItem()
        item.add(self.volume_button)
        item.set_tooltip_text(_("Volume"))
        self.player_toolbar.insert(item, -1)
        # Seekbar
        self.seekbar = Gtk.Scale(
            orientation=Gtk.Orientation.HORIZONTAL,
            adjustment=Gtk.Adjustment(value=0,
                                      lower=0,
                                      upper=1,
                                      step_increment=0.01,
                                      page_increment=0.05,
                                      page_size=0.05))

        self.seekbar.set_draw_value(False)
        self.seekbar.connect("change-value", self._on_seekbar_change_value)
        item = Gtk.ToolItem()
        item.set_expand(True)
        item.add(self.seekbar)
        self.player_toolbar.insert(item, -1)

    def _init_player_widgets(self):
        """Initialize the video player and related widgets."""
        vbox = gaupol.util.new_vbox(spacing=0)
        self.player = gaupol.VideoPlayer()
        aeidon.util.connect(self, "player", "state-changed")
        gaupol.util.pack_start_expand(vbox, self.player.widget)
        self._init_player_toolbar()
        gaupol.util.pack_start_fill(vbox, self.player_toolbar)
        gaupol.util.pack_start_expand(self.player_box, vbox)
        self.player_box.show_all()
        self.paned.add1(self.player_box)
        orientation = self.paned.get_orientation()
        size = (
            self.notebook.get_window().get_width()
            if orientation == Gtk.Orientation.HORIZONTAL
            else self.notebook.get_window().get_height())
        self.paned.set_position(int(size / 2))
        self.get_action("toggle-player").set_state(True)

    def _init_update_handlers(self):
        """Initialize timed updates of widgets."""
        while self._update_handlers:
            GLib.source_remove(self._update_handlers.pop())
        self._update_handlers = [
            GLib.timeout_add( 10, self._on_player_update_subtitle),
            GLib.timeout_add( 50, self._on_player_update_seekbar),
            GLib.timeout_add(100, self._on_player_update_volume),
        ]

    @aeidon.deco.export
    def load_video(self, path):
        """Load a video file."""
        page = self.get_current_page()
        page.project.video_path = path
        if self.player is None:
            self._init_player_widgets()
            self._init_cache_updates()
            self._init_update_handlers()
            self._update_subtitle_cache()
        else: # Player exists
            if self.player.is_playing():
                self.get_action("play-pause").activate()
            adjustment = self.seekbar.get_adjustment()
            adjustment.set_value(0)
            self.player.stop()
        self.player.set_path(path)
        # Playback initialization can fail, e.g. due to missing codecs,
        # in which case the player itself has shown an error dialog.
        if not self.player.ready: return
        self._update_languages_menu()
        self.update_gui()
        self.player.play()
        if not gaupol.conf.video_player.autoplay:
            self.player.pause()

    @aeidon.deco.export
    def _on_load_video_activate(self, *args):
        """Load a video file."""
        gaupol.util.set_cursor_busy(self.window)
        page = self.get_current_page()
        dialog = gaupol.VideoDialog(
            self.window, title=_("Load Video"), button_label=_("_Load"))
        if page.project.main_file is not None:
            directory = os.path.dirname(page.project.main_file.path)
            dialog.set_current_folder(directory)
        if page.project.video_path is not None:
            dialog.set_filename(page.project.video_path)
        gaupol.util.set_cursor_normal(self.window)
        response = gaupol.util.run_dialog(dialog)
        path = dialog.get_filename()
        dialog.destroy()
        if response != Gtk.ResponseType.OK: return
        self.load_video(path)

    @aeidon.deco.export
    def _on_play_pause_activate(self, *args):
        """Play or pause video."""
        if self.player.is_playing():
            self.player.pause()
        else: # Not playing.
            self.player.play()

    @aeidon.deco.export
    def _on_play_selection_activate(self, *args):
        """Play the selected subtitles."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        offset = gaupol.conf.video_player.context_length
        start = page.project.subtitles[rows[0]].start_seconds - offset
        end = page.project.subtitles[rows[-1]].end_seconds + offset
        self.player.play_segment(start, end)

    def _on_player_state_changed(self, player, state):
        """Update UI to match `state` of `player`."""
        self.play_button.set_icon_name(
            "media-playback-pause"
            if state == Gst.State.PLAYING
            else "media-playback-start")

    def _on_player_update_seekbar(self, data=None):
        """Update seekbar from video position."""
        if not self.player.ready:
            return True # to be called again.
        duration = self.player.get_duration(mode=None)
        position = self.player.get_position(mode=None)
        if duration is not None and position is not None:
            adjustment = self.seekbar.get_adjustment()
            adjustment.set_value(position/duration)
        return True # to be called again.

    def _on_player_update_subtitle(self, data=None):
        """Update subtitle overlay from video position."""
        if not self.player.ready:
            return True # to be called again.
        pos = self.player.get_position(aeidon.modes.SECONDS)
        if pos is None:
            return True # to be called again.
        subtitles = list(filter(lambda x: x[0] <= pos <= x[1], self._cache))
        if subtitles:
            text = subtitles[-1][2]
            if text != self.player.subtitle_text_raw:
                self.player.subtitle_text = text
        else:
            if self.player.subtitle_text:
                self.player.subtitle_text = ""
        return True # to be called again.

    def _on_player_update_volume(self, data=None):
        """Update volume from player."""
        if not self.player.ready:
            return True # to be called again.
        self.volume_button.set_value(self.player.volume)
        return True # to be called again.

    @aeidon.deco.export
    def _on_seek_backward_activate(self, *args):
        """Seek backward."""
        pos = self.player.get_position(aeidon.modes.SECONDS)
        if pos is None: return
        pos = pos - gaupol.conf.video_player.seek_length
        pos = max(pos, 0)
        self.player.seek(pos)

    @aeidon.deco.export
    def _on_seek_forward_activate(self, *args):
        """Seek forward."""
        position = self.player.get_position(aeidon.modes.SECONDS)
        duration = self.player.get_duration(aeidon.modes.SECONDS)
        if position is None: return
        if duration is None: return
        position = position + gaupol.conf.video_player.seek_length
        position = min(position, duration)
        self.player.seek(position)

    @aeidon.deco.export
    def _on_seek_next_activate(self, *args):
        """Seek to the start of the next subtitle."""
        pos = self.player.get_position(aeidon.modes.SECONDS)
        if pos is None: return
        subtitles = list(filter(lambda x: x[0] > pos + 0.001, self._cache))
        if not subtitles: return
        self.player.seek(subtitles[0][0])

    @aeidon.deco.export
    def _on_seek_previous_activate(self, *args):
        """Seek to the start of the previous subtitle."""
        pos = self.player.get_position(aeidon.modes.SECONDS)
        if pos is None: return
        subtitles = list(filter(lambda x: x[1] < pos - 0.001, self._cache))
        if not subtitles: return
        self.player.seek(subtitles[-1][0])

    @aeidon.deco.export
    def _on_seek_selection_end_activate(self, *args):
        """Seek to the end of selection."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        pos = page.project.subtitles[rows[-1]].end_seconds
        offset = gaupol.conf.video_player.context_length
        self.player.seek(max(pos - offset, 0))

    @aeidon.deco.export
    def _on_seek_selection_start_activate(self, *args):
        """Seek to the start of selection."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        pos = page.project.subtitles[rows[0]].start_seconds
        offset = gaupol.conf.video_player.context_length
        self.player.seek(max(pos - offset, 0))

    def _on_seekbar_change_value(self, seekbar, scroll, value, data=None):
        """Seek to specified position in video."""
        self.player.subtitle_text = ""
        duration = self.player.get_duration(aeidon.modes.SECONDS)
        if duration is None: return
        self.player.seek(value * duration)

    @aeidon.deco.export
    def _on_set_audio_language_activate(self, action, parameter):
        """Set the audio language to use."""
        # Avoid freeze by pausing prior and playing after.
        # https://github.com/otsaloma/gaupol/issues/58
        self.player.pause()
        index = int(parameter.get_string())
        self.player.audio_track = index
        self._update_languages_menu()
        self.player.play()

    def _on_volume_button_value_changed(self, button, value):
        """Update video player volume."""
        self.player.volume = value
        self.update_gui()

    @aeidon.deco.export
    def _on_volume_down_activate(self, *args):
        """Decrease volume."""
        self.player.volume = self.player.volume - 0.05
        self.volume_button.set_value(self.player.volume)
        self.update_gui()

    @aeidon.deco.export
    def _on_volume_up_activate(self, *args):
        """Increase volume."""
        self.player.volume = self.player.volume + 0.05
        self.volume_button.set_value(self.player.volume)
        self.update_gui()

    def _update_languages_menu(self):
        """Update the audio language selection menu."""
        menu = self.get_menubar_section("audio-languages-placeholder")
        menu.remove_all()
        tracks = self.player.get_audio_infos()
        for i, track in enumerate(tracks):
            title = track.title or _("Track {:d}").format(i + 1)
            lang = None
            if track.language_code is not None:
                lang = GstTag.tag_get_language_name(track.language_code)
            if lang is None:
                lang = track.language_name
            name = title if lang is None else "{} - [{}]".format(title, lang.title())
            action = "win.set-audio-language::{:d}".format(i)
            menu.append(name, action)
            if i == self.player.audio_track:
                action = self.get_action("set-audio-language")
                action.set_state(str(i))

    def _update_subtitle_cache(self, *args, **kwargs):
        """Update subtitle position and text cache."""
        page = self.get_current_page()
        if self.player is None or page is None:
            return self._clear_subtitle_cache()
        self._cache = [(x.start_seconds, x.end_seconds, x.main_text)
                       for x in page.project.subtitles]



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for informing that preview failed."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk


class PreviewErrorDialog(Gtk.MessageDialog):

    """Dialog for informing that preview failed."""

    def __init__(self, parent, output):
        """Initialize a :class:`PreviewErrorDialog` instance."""
        GObject.GObject.__init__(self,
                                 message_type=Gtk.MessageType.ERROR,
                                 text=_("Preview in video player failed"),
                                 secondary_text=_("There is probably a problem with either the video file or the video player. See the preferences dialog to choose the video player or to customize the command."))

        self._text_view = Gtk.TextView()
        self._init_dialog(parent)
        self._init_text_view(output)

    def _init_dialog(self, parent):
        """Initialize the dialog."""
        self.add_button(_("_Close"), Gtk.ResponseType.CLOSE)
        self.set_default_response(Gtk.ResponseType.CLOSE)
        self.set_transient_for(parent)
        self.set_modal(True)

    def _init_text_view(self, output):
        """Initialize the text view."""
        self._text_view.set_wrap_mode(Gtk.WrapMode.WORD)
        self._text_view.set_editable(False)
        self._text_view.set_cursor_visible(False)
        self._text_view.set_accepts_tab(False)
        self._text_view.set_left_margin(6)
        self._text_view.set_right_margin(6)
        with aeidon.util.silent(AttributeError):
            # Available since GTK 3.18.
            self._text_view.set_top_margin(6)
            self._text_view.set_bottom_margin(6)
        text_buffer = self._text_view.get_buffer()
        text_buffer.set_text(output)
        gaupol.style.use_font(self._text_view, "monospace")
        scroller = Gtk.ScrolledWindow()
        scroller.set_policy(*((Gtk.PolicyType.AUTOMATIC,)*2))
        scroller.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scroller.add(self._text_view)
        box = self.get_message_area()
        gaupol.util.pack_start_expand(box, scroller)
        gaupol.util.scale_to_content(self._text_view,
                                     min_nchar=60,
                                     max_nchar=100,
                                     min_nlines=10,
                                     max_nlines=25,
                                     font="monospace")

        box.show_all()



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Simple subtitle data editing actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

from gi.repository import Gtk

class EditPreferencesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "edit-preferences")
        self.action_group = "safe"

class EditNextValueAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "edit-next-value")
        self.accelerators = ["space"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        row, col = page.view.get_focus()
        aeidon.util.affirm(not None in (row, col))
        aeidon.util.affirm(row < len(page.project.subtitles) - 1)
        column = page.view.get_column(col)
        renderer = column.get_cells()[0]
        aeidon.util.affirm(renderer.props.mode ==
                           Gtk.CellRendererMode.EDITABLE)

class EditValueAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "edit-value")
        self.accelerators = ["Return"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        row, col = page.view.get_focus()
        aeidon.util.affirm(not None in (row, col))
        column = page.view.get_column(col)
        renderer = column.get_cells()[0]
        aeidon.util.affirm(renderer.props.mode ==
                           Gtk.CellRendererMode.EDITABLE)

class EndEarlierAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "end-earlier")
        self.accelerators = ["E"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(len(selected_rows) == 1)

class EndLaterAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "end-later")
        self.accelerators = ["R"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(len(selected_rows) == 1)

class ExtendSelectionToBeginningAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "extend-selection-to-beginning")
        self.accelerators = ["<Shift><Control>Home"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(selected_rows)

class ExtendSelectionToEndAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "extend-selection-to-end")
        self.accelerators = ["<Shift><Control>End"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(selected_rows)

class InsertSubtitleAtVideoPositionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "insert-subtitle-at-video-position")
        self.accelerators = ["J"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class InsertSubtitlesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "insert-subtitles")
        self.accelerators = ["I"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        if page.project.subtitles:
            aeidon.util.affirm(selected_rows)

class InvertSelectionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "invert-selection")
        self.accelerators = ["<Control>J"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)

class MergeSubtitlesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "merge-subtitles")
        self.accelerators = ["M"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(len(selected_rows) > 1)
        block = list(range(selected_rows[0], selected_rows[-1] + 1))
        aeidon.util.affirm(list(selected_rows) == block)

class RedoActionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "redo-action")
        self.accelerators = ["<Shift><Control>Z"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.can_redo())

class RemoveSubtitlesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "remove-subtitles")
        self.accelerators = ["Delete"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)

class SelectAllAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "select-all")
        self.accelerators = ["<Control>A"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)

class SelectNextFromVideoPositionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "select-next-from-video-position")
        self.accelerators = ["<Control>U"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SelectPreviousFromVideoPositionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "select-previous-from-video-position")
        self.accelerators = ["<Control>Y"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)

class SetEndFromVideoPositionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "set-end-from-video-position")
        self.accelerators = ["K"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)
        aeidon.util.affirm(len(selected_rows) == 1)

class SetStartFromVideoPositionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "set-start-from-video-position")
        self.accelerators = ["U"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(application.player is not None)
        aeidon.util.affirm(application.player.ready)
        aeidon.util.affirm(len(selected_rows) == 1)

class SplitSubtitleAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "split-subtitle")
        self.accelerators = ["S"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(len(selected_rows) == 1)

class StartEarlierAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "start-earlier")
        self.accelerators = ["Q"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(len(selected_rows) == 1)

class StartLaterAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "start-later")
        self.accelerators = ["W"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.subtitles)
        aeidon.util.affirm(len(selected_rows) == 1)

class UndoActionAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "undo-action")
        self.accelerators = ["<Control>Z"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.can_undo())

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Opening subtitle files and creating new projects."""

import aeidon
import gaupol
import os

from aeidon.i18n   import _
from gi.repository import Gtk


class OpenAgent(aeidon.Delegate):

    """Opening subtitle files and creating new projects."""

    @aeidon.deco.export
    def add_page(self, page):
        """Add `page` to the application."""
        self.pages.append(page)
        page.connect("close-request", self._on_page_close_request)
        page.project.connect("action-done", self._on_project_action_done)
        page.project.connect("action-redone", self._on_project_action_redone)
        page.project.connect("action-undone", self._on_project_action_undone)
        callback = self._on_tab_widget_button_press_event
        page.tab_widget.connect("button-press-event", callback, page)
        self.connect_view_signals(page.view)
        page.project.clipboard.set_texts(self.clipboard.get_texts())
        scroller = Gtk.ScrolledWindow()
        policy = Gtk.PolicyType.AUTOMATIC
        scroller.set_policy(policy, policy)
        scroller.add(page.view)
        self.notebook.append_page(scroller, page.tab_widget)
        self.notebook.set_tab_reorderable(scroller, True)
        self.notebook.child_set_property(scroller, "tab-expand", True)
        self.notebook.child_set_property(scroller, "tab-fill", True)
        self.notebook.show_all()
        self.set_current_page(page)
        self.emit("page-added", page)

    @aeidon.deco.export
    def add_to_recent_files(self, path, format, doc):
        """Add `path` to recent files managed by the recent manager."""
        # XXX: The group field is not available for Python,
        # we cannot differentiate between main and translation files.
        # https://bugzilla.gnome.org/show_bug.cgi?id=695970
        uri = aeidon.util.path_to_uri(path)
        recent = Gtk.RecentData()
        recent.mime_type = format.mime_type
        recent.app_name = "gaupol"
        recent.app_exec = "gaupol %F"
        self.recent_manager.add_full(uri, recent)

    @aeidon.deco.export
    def append_file(self, path, encoding=None):
        """Append subtitles from file at `path` to the current project."""
        encodings = self._get_encodings(encoding)
        doc = aeidon.documents.MAIN
        temp = self._open_file(path, encodings, doc, check_open=False)
        gaupol.util.set_cursor_busy(self.window)
        current = self.get_current_page()
        offset = current.project.subtitles[-1].end
        temp.project.shift_positions(None, offset)
        rows = self._append_subtitles(current, temp.project.subtitles)
        amount = len(rows)
        current.view.set_focus(rows[0], None)
        current.view.select_rows(rows)
        current.view.scroll_to_row(rows[0])
        basename = temp.get_main_basename()
        message = _('Appended {amount:d} subtitles from "{basename}"')
        self.flash_message(message.format(**locals()))
        gaupol.util.set_cursor_normal(self.window)

    def _append_subtitles(self, page, subtitles):
        """Append `subtitles` to `page` and return new indices."""
        n = len(page.project.subtitles)
        indices = list(range(n, n + len(subtitles)))
        page.project.block("action-done")
        page.project.insert_subtitles(indices, subtitles)
        page.project.set_action_description(
            aeidon.registers.DO, _("Appending file"))
        page.project.unblock("action-done")
        return tuple(indices)

    def _check_file_exists(self, path):
        """Raise :exc:`gaupol.Default` if no file at `path`."""
        gaupol.util.raise_default(not os.path.isfile(path))

    def _check_file_not_open(self, path):
        """Raise :exc:`gaupol.Default` if file at `path` already open."""
        for page in self.pages:
            files = [page.project.main_file, page.project.tran_file]
            paths = [x.path for x in files if x]
            if not path in paths: continue
            self.set_current_page(page)
            message = _('File "{}" is already open')
            self.flash_message(message.format(os.path.basename(path)))
            raise gaupol.Default

    def _check_file_size(self, path):
        """Raise :exc:`gaupol.Default` if size of file at `path` too large."""
        size_mb = os.stat(path).st_size / 1048576
        if size_mb <= 1: return
        basename = os.path.basename(path)
        self._show_size_warning_dialog(basename, size_mb)

    def _check_sort_count(self, path, sort_count):
        """Raise :exc:`gaupol.Default` if `sort_count` too large."""
        if sort_count <= 0: return
        basename = os.path.basename(path)
        self._show_sort_warning_dialog(basename, sort_count)

    @aeidon.deco.export
    def connect_view_signals(self, view):
        """Connect to signals emitted by `view`."""
        view.connect_selection_changed(self._on_view_selection_changed)
        view.connect_after("move-cursor", self._on_view_move_cursor)
        view.connect("button-press-event", self._on_view_button_press_event)
        for column in view.get_columns():
            renderer = column.get_cells()[0]
            callback = self._on_view_renderer_edited
            renderer.connect("edited", callback, column)
            callback = self._on_view_renderer_editing_started
            renderer.connect("editing-started", callback, column)
            callback = self._on_view_renderer_editing_canceled
            renderer.connect("editing-canceled", callback, column)
            button = column.get_widget().get_ancestor(Gtk.Button)
            callback = self._on_view_header_button_press_event
            button.connect("button-press-event", callback)

    def _get_encodings(self, first=None):
        """Return a sequence of encodings to try when opening files."""
        encodings = [first]
        if gaupol.conf.encoding.try_locale:
            encoding = aeidon.encodings.get_locale_code()
            encodings.append(encoding)
        encodings += gaupol.conf.encoding.fallback
        try_auto = gaupol.conf.encoding.try_auto
        if try_auto and aeidon.util.chardet_available():
            encodings.append("auto")
        encodings = list(filter(None, encodings))
        encodings = encodings or ["utf_8"]
        return tuple(aeidon.util.get_unique(encodings))

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_append_file_activate(self, *args):
        """Append subtitles from file to the current project."""
        gaupol.util.set_cursor_busy(self.window)
        dialog = gaupol.AppendDialog(self.window)
        gaupol.util.set_cursor_normal(self.window)
        response = gaupol.util.run_dialog(dialog)
        paths = dialog.get_filenames()
        encoding = dialog.get_encoding()
        dialog.destroy()
        if response != Gtk.ResponseType.OK: return
        if not paths: return
        gaupol.util.iterate_main()
        self.append_file(paths[0], encoding)

    @aeidon.deco.export
    def _on_new_project_activate(self, *args):
        """Create a new project."""
        if gaupol.fields.TRAN_TEXT in gaupol.conf.editor.visible_fields:
            gaupol.conf.editor.visible_fields.remove(gaupol.fields.TRAN_TEXT)
        page = gaupol.Page(next(self.counter))
        page.project.insert_subtitles((0,), register=None)
        self.add_page(page)

    @aeidon.deco.export
    def _on_notebook_drag_data_received(self, notebook, context, x, y,
                                        selection_data, info, time):

        """Open main files from dragged URIs."""
        uris = selection_data.get_uris()
        paths = list(map(aeidon.util.uri_to_path, uris))
        videos = list(filter(aeidon.util.is_video_file, paths))
        subtitles = list(set(paths) - set(videos))
        self.open_main(subtitles)
        if self.get_current_page() and len(videos) == 1:
            self.load_video(videos[0])

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_open_main_files_activate(self, *args):
        """Open main files."""
        doc = aeidon.documents.MAIN
        paths, encoding = self._select_files(_("Open"), doc)
        self.open_main(paths, encoding)

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_open_translation_file_activate(self, *args):
        """Open a translation file."""
        page = self.get_current_page()
        if page.project.tran_changed:
            self._show_translation_warning_dialog(page)
        doc = aeidon.documents.TRAN
        paths, encoding = self._select_files(_("Open Translation"), doc)
        self.open_translation(paths[0], encoding)

    @aeidon.deco.export
    def _on_select_video_file_activate(self, *args):
        """Select a video file."""
        gaupol.util.set_cursor_busy(self.window)
        page = self.get_current_page()
        path = page.project.video_path
        title = _("Select Video")
        label = _("_Select")
        dialog = gaupol.VideoDialog(self.window, title, label)
        if page.project.main_file is not None:
            directory = os.path.dirname(page.project.main_file.path)
            dialog.set_current_folder(directory)
        if page.project.video_path is not None:
            dialog.set_filename(page.project.video_path)
        gaupol.util.set_cursor_normal(self.window)
        response = gaupol.util.run_dialog(dialog)
        path = dialog.get_filename()
        dialog.destroy()
        if response != Gtk.ResponseType.OK: return
        page.project.video_path = path
        self.update_gui()

    @aeidon.deco.export
    def _on_split_project_activate(self, *args):
        """Split the current project in two."""
        gaupol.util.flash_dialog(gaupol.SplitDialog(self.window, self))

    def _open_file(self, path, encodings, doc, check_open=True):
        """Open file at `path` and return corresponding page if successful."""
        self._check_file_exists(path)
        if check_open:
            self._check_file_not_open(path)
        self._check_file_size(path)
        basename = os.path.basename(path)
        page = (gaupol.Page() if doc == aeidon.documents.MAIN
                else self.get_current_page())
        for encoding in encodings:
            with aeidon.util.silent(UnicodeError):
                n = self._try_open_file(page, doc, path, encoding)
                self._check_sort_count(path, n)
                return page
        # Report if all codecs failed to decode file.
        self._show_encoding_error_dialog(basename)
        raise gaupol.Default

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def open_main(self, path, encoding=None):
        """Open file at `path` as a main file."""
        if gaupol.fields.TRAN_TEXT in gaupol.conf.editor.visible_fields:
            gaupol.conf.editor.visible_fields.remove(gaupol.fields.TRAN_TEXT)
        encodings = self._get_encodings(encoding)
        gaupol.util.set_cursor_busy(self.window)
        for path in aeidon.util.flatten([path]):
            try:
                # Skip files that are already open,
                # but show a status message when that happens.
                self._check_file_not_open(path)
            except gaupol.Default:
                continue
            try:
                page = self._open_file(path, encodings, aeidon.documents.MAIN)
            except gaupol.Default:
                gaupol.util.set_cursor_normal(self.window)
                raise # gaupol.Default
            self.add_page(page)
            format = page.project.main_file.format
            self.add_to_recent_files(path, format, aeidon.documents.MAIN)
            # Refresh view to get row heights etc. correct.
            page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        gaupol.util.set_cursor_normal(self.window)
        self.update_gui()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def open_translation(self, path, encoding=None, align_method=None):
        """Open file at `path` as a translation file."""
        if align_method is not None:
            gaupol.conf.file.align_method = align_method
        encodings = self._get_encodings(encoding)
        page = self._open_file(path, encodings, aeidon.documents.TRAN)
        gaupol.util.set_cursor_busy(self.window)
        col = page.view.columns.TRAN_TEXT
        if not page.view.get_column(col).get_visible():
            self.get_column_action(gaupol.fields.TRAN_TEXT).activate()
        format = page.project.tran_file.format
        self.add_to_recent_files(path, format, aeidon.documents.TRAN)
        gaupol.util.set_cursor_normal(self.window)

    def _select_files(self, title, doc):
        """Show a :class:`gaupol.OpenDialog` to select files."""
        gaupol.util.set_cursor_busy(self.window)
        dialog = gaupol.OpenDialog(self.window, title, doc)
        page = self.get_current_page()
        if page is not None and page.project.main_file is not None:
            directory = os.path.dirname(page.project.main_file.path)
            dialog.set_current_folder(directory)
        gaupol.util.set_cursor_normal(self.window)
        response = gaupol.util.run_dialog(dialog)
        paths = dialog.get_filenames()
        encoding = dialog.get_encoding()
        dialog.destroy()
        gaupol.util.raise_default(response != Gtk.ResponseType.OK)
        gaupol.util.iterate_main()
        return paths, encoding

    def _show_encoding_error_dialog(self, basename):
        """Show an error dialog after failing to decode file."""
        title = _('Failed to decode file "{}" with all attempted codecs').format(basename)
        message = _("Please try to open the file with a different character encoding.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_format_error_dialog(self, basename):
        """Show an error dialog after failing to recognize file format."""
        title = _('Failed to recognize format of file "{}"').format(basename)
        message = _("Please check that the file you are trying to open is a subtitle file of a format supported by Gaupol.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_io_error_dialog(self, basename, message):
        """Show an error dialog after failing to read file."""
        title = _('Failed to open file "{}"').format(basename)
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_parse_error_dialog(self, basename, format):
        """Show an error dialog after failing to parse file."""
        title = _('Failed to parse file "{}"').format(basename)
        message = _("Please check that the file you are trying to open is a valid {} file.").format(format.label)
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_size_warning_dialog(self, basename, size):
        """Show a warning dialog when trying to open a large file."""
        title = _('Open abnormally large file "{}"?').format(basename)
        message = _("Size of the file is {:.1f} MB, which is abnormally large for a text-based subtitle file. Please, check that you are not trying to open a binary file.").format(size)
        dialog = gaupol.WarningDialog(self.window, title, message)
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.NO)
        dialog.add_button(_("_Open"), Gtk.ResponseType.YES)
        dialog.set_default_response(Gtk.ResponseType.NO)
        response = gaupol.util.flash_dialog(dialog)
        gaupol.util.raise_default(response != Gtk.ResponseType.YES)

    def _show_sort_warning_dialog(self, basename, count):
        """Show a warning dialog when subtitles have been sorted."""
        title = _('Open unsorted file "{}"?').format(basename)
        message = _("The order of {:d} subtitles needs to be changed. If {:d} sounds like a lot, the file may be erroneously composed.")
        message = message.format(count, count)
        dialog = gaupol.WarningDialog(self.window, title, message)
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.NO)
        dialog.add_button(_("_Open"), Gtk.ResponseType.YES)
        dialog.set_default_response(Gtk.ResponseType.YES)
        response = gaupol.util.flash_dialog(dialog)
        gaupol.util.raise_default(response != Gtk.ResponseType.YES)

    def _show_translation_warning_dialog(self, page):
        """Show a warning dialog if opening a new translation file."""
        title = _('Save changes to translation document "{}" before opening a new one?').format(page.get_translation_basename())
        message = _("If you don't save, changes will be permanently lost.")
        dialog = gaupol.WarningDialog(self.window, title, message)
        dialog.add_button(_("Open _Without Saving"), Gtk.ResponseType.NO)
        dialog.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        dialog.add_button(_("_Save"), Gtk.ResponseType.YES)
        dialog.set_default_response(Gtk.ResponseType.YES)
        response = gaupol.util.flash_dialog(dialog)
        if response == Gtk.ResponseType.YES:
            return self.save_translation(page)
        gaupol.util.raise_default(response != Gtk.ResponseType.NO)

    def _try_open_file(self, page, doc, path, encoding, **kwargs):
        """Try to open file at `path` and return subtitle sort count."""
        if encoding == "auto":
            encoding = aeidon.encodings.detect(path)
            if encoding is None: raise UnicodeError
        kwargs["align_method"] = gaupol.conf.file.align_method
        basename = os.path.basename(path)
        try:
            return page.project.open(doc, path, encoding, **kwargs)
        except aeidon.FormatError:
            self._show_format_error_dialog(basename)
        except IOError as error:
            self._show_io_error_dialog(basename, str(error))
        except aeidon.ParseError:
            bom_encoding = aeidon.encodings.detect_bom(path)
            encoding = bom_encoding or encoding
            with aeidon.util.silent(Exception):
                format = aeidon.util.detect_format(path, encoding)
            self._show_parse_error_dialog(basename, format)
        raise gaupol.Default



# -*- coding: utf-8 -*-

# Copyright (C) 2006 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Observable dictionary with attribute access to keys."""

import aeidon

__all__ = ("AttributeDictionary",)


class AttributeDictionary(aeidon.Observable):

    """
    Observable dictionary with attribute access to keys.

    :class:`AttributeDictionary` is initialized from a root dictionary,
    which is kept in sync with attribute values. This allows convenient
    attribute access to dictionary keys and notifications of changes
    via the :class:`aeidon.Observable` interface.
    """

    def __init__(self, root):
        """Initialize an :class:`AttributeDictionary` instance."""
        aeidon.Observable.__init__(self)
        self._root = root
        self.update(root)

    def add_attribute(self, name, value):
        """Add instance attribute and corresponding root dictionary key."""
        self._root[name] = value
        # In the case of dictionaries, set the original dictionary
        # to the root dictionary, but instantiate an AttributeDictionary
        # for use as the corresponding attribute.
        if isinstance(value, dict):
            value = AttributeDictionary(value)
        setattr(self, name, value)
        self.connect("notify::{}".format(name), self._on_notify, name)

    def extend(self, root):
        """Add new values from another root dictionary."""
        for name, value in root.items():
            if not hasattr(self, name):
                self.add_attribute(name, value)
        for name, value in root.items():
            if isinstance(value, dict):
                getattr(self, name).extend(value)

    def _on_notify(self, obj, value, name):
        """Synchronize changed attribute value with root dictionary."""
        self._root[name] = value

    def update(self, root):
        """Update values from another root dictionary."""
        self.extend(root)
        for name, value in root.items():
            if not isinstance(value, dict):
                setattr(self, name, value)
        for name, value in root.items():
            if isinstance(value, dict):
                getattr(self, name).update(value)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""String and regular expression finder and replacer."""

import re

__all__ = ("Finder",)


class Finder:

    """
    String and regular expression finder and replacer.

    :ivar ignore_case: ``True`` to ignore case when finding matches
    :ivar match: Regular expression object for the latest match of pattern
    :ivar match_span: Tuple of start and end position for match
    :ivar pattern: String or regular expression object to find
    :ivar pos: Current offset from the beginning of the text
    :ivar replacement: Plain- or regular expression replacement string
    :ivar text: Target text to find matches of pattern in
    """

    def __init__(self):
        """Initialize a :class:`Finder` instance."""
        self.ignore_case = False
        self.match = None
        self.match_span = None
        self.pattern = None
        self.pos = None
        self.replacement = None
        self.text = None

    def next(self):
        """
        Find the next match of pattern.

        Raise :exc:`StopIteration` if no next match found.
        Return tuple of match start, end position.
        """
        if self.pos is None:
            # Start new search from beginning.
            self.pos = 0
        if isinstance(self.pattern, str):
            text = self.text
            pattern = self.pattern
            if self.ignore_case:
                text = text.lower()
                pattern = pattern.lower()
            try:
                index = text.index(pattern, self.pos)
            except ValueError:
                raise StopIteration
            self.match_span = (index, index + len(pattern))
        else: # Regular expression
            match = self.pattern.search(self.text, self.pos)
            if match is None:
                raise StopIteration
            # Avoid getting stuck with zero-length regular expressions.
            if match.span() == self.match_span == (self.pos, self.pos):
                if self.pos == len(self.text):
                    raise StopIteration
                self.pos += 1
                return self.next()
            self.match = match
            self.match_span = match.span()
        self.pos = self.match_span[1]
        return self.match_span

    def previous(self):
        """
        Find the previous match of pattern.

        Raise :exc:`StopIteration` if no previous match found.
        Return tuple of match start, end position.
        """
        if self.pos is None:
            # Start new search from end.
            self.pos = len(self.text)
        if isinstance(self.pattern, str):
            text = self.text
            pattern = self.pattern
            if self.ignore_case:
                text = text.lower()
                pattern = pattern.lower()
            try:
                index = text.rindex(pattern, 0, self.pos)
            except ValueError:
                raise StopIteration
            self.match_span = (index, index + len(pattern))
        else: # Regular expression
            iterator = self.pattern.finditer(self.text)
            match = None
            while True:
                try:
                    candidate = next(iterator)
                    if candidate.end() > self.pos:
                        raise StopIteration
                except StopIteration:
                    break
                match = candidate
            if match is None:
                raise StopIteration
            # Avoid getting stuck with zero-length regular expressions.
            if match.span() == self.match_span == (self.pos, self.pos):
                if self.pos == 0:
                    raise StopIteration
                self.pos -= 1
                return self.previous()
            self.match = match
            self.match_span = match.span()
        self.pos = self.match_span[0]
        return self.match_span

    def replace(self, next=True):
        """
        Replace the current match of pattern.

        `next` should be ``True`` to finish at end of match, ``False`` for
        beginning. Raise :exc:`re.error` if bad replacement.
        """
        a, z = self.match_span
        orig_length = len(self.text)
        replacement = self.replacement
        if not isinstance(self.pattern, str):
            replacement = self.match.expand(self.replacement)
        self.text = self.text[:a] + replacement + self.text[z:]
        shift = len(self.text) - orig_length
        self.pos = (z + shift) if next else a
        # Adapt match span to new text length to avoid
        # getting stuck with zero-length regular expressions.
        if next and (self.match_span[0] == self.match_span[1]):
            self.match_span = (self.pos, self.pos)

    def replace_all(self):
        """
        Replace all occurences of pattern.

        Raise :exc:`re.error` if bad replacement.
        Return the amount of substitutions made.
        """
        self.pos = 0
        self.match = None
        self.match_span = None
        count = 0
        while True:
            try:
                self.next()
            except StopIteration:
                self.pos = len(self.text)
                self.match_span = None
                break
            self.replace()
            count += 1
        return count

    def set_regex(self, pattern, flags=re.DOTALL|re.MULTILINE):
        """
        Set and use regular expression as pattern.

        The default value for `flags` is ``DOTALL`` and ``MULTILINE``.
        ``IGNORECASE`` is automatically added to flags if :attr:`ignore_case`
        is ``True``. Raise :exc:`re.error` if bad pattern.
        """
        if self.ignore_case:
            flags = flags | re.IGNORECASE
        self.pattern = re.compile(pattern, flags)

    def set_text(self, text):
        """Set the target text to search in and reset position."""
        self.text = text
        self.match = None
        self.match_span = None
        self.pos = None



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base class for observable objects."""

import aeidon

__all__ = ("Observable",)


class Observable:

    """
    Base class for observable objects.

    :cvar signals: Tuple of emittable signals added automatically

    In addition to the signals defined in :attr:`signals`, all public instance
    variables have a ``notify::NAME`` signal generated automatically based on
    the ``NAME`` of the variable. ``notify::NAME`` signals will be emitted
    whenever the value of the corresponding instance variable changes.

    Notify signals will be emitted for mutable variables as well, which means
    that care should be taken not to emit thousands of signals when
    e.g. appending one-by-one to a large list. :meth:`freeze_notify` and
    :meth:`thaw_notify` will queue notify signals and emit only one of each
    once thawed.

    The Observable philosophy and API is highly inspired by GObject_.

    .. _GObject: https://developer.gnome.org/gobject/
    """

    __slots__ = (
        "_blocked_signals",
        "_blocked_state",
        "_notify_frozen",
        "_notify_queue",
        "_signal_handlers",
    )

    signals = ()

    def __init__(self):
        """Initialize an :class:`Observable` instance."""
        self._blocked_signals = []
        self._blocked_state = False
        self._notify_frozen = False
        self._notify_queue = []
        self._signal_handlers = {}
        for signal in self.signals:
            self._add_signal(signal)

    def __setattr__(self, name, value):
        """Set value of observable attribute."""
        if (name in self.__slots__) or name.startswith("_"):
            return object.__setattr__(self, name, value)
        value = self._validate(name, value)
        signal = "notify::{}".format(name)
        if not signal in self._signal_handlers:
            self._add_signal(signal)
            return object.__setattr__(self, name, value)
        return_value = object.__setattr__(self, name, value)
        self.emit(signal, value)
        return return_value

    def _add_signal(self, signal):
        """Add `signal` to the list of signals emitted."""
        self._signal_handlers[signal] = []

    def block(self, signal):
        """
        Block all emissions of `signal`.

        Return ``False`` if already blocked, otherwise ``True``.
        """
        if not signal in self._blocked_signals:
            self._blocked_signals.append(signal)
            return True
        return False

    def block_all(self):
        """
        Block all emissions of all signals.

        Return ``False`` if already blocked, otherwise ``True``.
        """
        if not self._blocked_state:
            self._blocked_state = True
            return True
        return False

    def connect(self, signal, method, *args):
        """Register to receive notifications of ``signal``."""
        self._signal_handlers[signal].append((method, args))

    def disconnect(self, signal, method):
        """Remove registration to receive notifications of ``signal``."""
        for i in reversed(range(len(self._signal_handlers[signal]))):
            if self._signal_handlers[signal][i][0] == method:
                self._signal_handlers[signal].pop(i)

    def emit(self, signal, *args):
        """Send notification of ``signal`` to all registered observers."""
        if signal.startswith("notify::") and self._notify_frozen:
            if not signal in self._notify_queue:
                self._notify_queue.append(signal)
            return
        if (not self._blocked_state and
            not signal in self._blocked_signals):
            if signal.startswith("notify::"):
                name = signal.replace("notify::", "")
                args = (getattr(self, name),)
            for method, data in self._signal_handlers[signal]:
                method(*((self,) + args + data))

    def freeze_notify(self):
        """
        Queue notify signals instead of emitting them.

        Return ``False`` if already frozen, otherwise ``True``.
        """
        if not self._notify_frozen:
            self._notify_frozen = True
            return True
        return False

    def notify(self, name):
        """Emit notification signal for variable."""
        return self.emit("notify::{}".format(name))

    def thaw_notify(self, do=True):
        """
        Emit all queued notify signals and queue no more.

        The optional `do` keyword argument should be the return value from
        :meth:`freeze_notify` to avoid problems with nested functions where
        notifications were frozen at a higher level. If `do` is ``False``,
        nothing will be done.

        Return ``False`` if already thawed, otherwise ``True``.
        """
        if do and self._notify_frozen:
            self._notify_frozen = False
            for signal in self._notify_queue:
                name = signal.replace("notify::", "")
                self.emit(signal, getattr(self, name))
            self._notify_queue = []
            return True
        return False

    def unblock(self, signal, do=True):
        """
        Unblock all emissions of `signal`.

        The optional `do` keyword argument should be the return value from
        :meth:`block` to avoid problems with nested functions where signals
        were blocked at a higher level. If `do` is ``False``, nothing will be
        done.

        Return ``False`` if already unblocked, otherwise ``True``.
        """
        if do and (signal in self._blocked_signals):
            self._blocked_signals.remove(signal)
            return True
        return False

    def unblock_all(self, do=True):
        """
        Unblock all emissions of all signals.

        The optional `do` keyword argument should be the return value from
        :meth:`block_all` to avoid problems with nested functions where signals
        were blocked at a higher level. If `do` is ``False``, nothing will be
        done.

        Return ``False`` if already unblocked, otherwise ``True``.
        """
        if do and self._blocked_state:
            self._blocked_state = False
            return True
        return False

    def _validate(self, name, value):
        """Return `value` or an observable version if `value` is mutable."""
        args = (value, self, name)
        if isinstance(value, dict):
            return aeidon.ObservableDict(*args)
        if isinstance(value, list):
            return aeidon.ObservableList(*args)
        if isinstance(value, set):
            return aeidon.ObservableSet(*args)
        return value



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Lists of named constants with integer values.

:class:`EnumerationItem`s are subclassed from :class:`int`, so they are actual
integers. The string value of an item will be the name that it was defined with
in its set. New items can always be added to an enumeration.
"""

__all__ = ("EnumerationItem", "Enumeration",)


class EnumerationItem(int):

    """
    Named constant with an integer value.

    :class:`EnumerationItem` can be instantiated without any specified values
    and users need not bother with them. Instantiation with correct values
    harmonous with the rest of the items should be left up to the parent list.
    """

    def __new__(cls, value=0, name="", parent=None):
        """Return integer instance with additional attributes."""
        instance = int.__new__(cls, value)
        instance.name = name
        instance.parent = parent
        return instance

    def __bool__(self):
        """For consistency, always return ``True``."""
        return True

    def __str__(self):
        """Return name as the string representation."""
        return self.name


class Enumeration(list):

    """
    List of named constants with integer values.

    :class:`Enumeration` is an actual :class:`list` where enumeration items are
    stored as both list items and instance attributes. New items should be
    added by setting an instance attribute.

    Typical use to create a new enumeration would be something like::

        fruits = aeidon.Enumeration()
        fruits.APPLE = aeidon.EnumerationItem()
        fruits.MANGO = aeidon.EnumerationItem()
        fruits.APPLE.size = 10
        fruits.MANGO.size = 20

    Note that there is no finalization of an enumeration. New items can always
    be added just by assigning a new attribute to the enumeration. Likewise,
    existing items can always be removed using :func:`delattr`.
    """

    NONE = None

    def __contains__(self, item):
        """Return ``True`` if enumeration contains `item`."""
        return list.__contains__(self, item)

    def __delattr__(self, name):
        """Delete enumeration item and attribute."""
        list.remove(self, getattr(self, name))
        return object.__delattr__(self, name)

    def find_item(self, name, value):
        """Return the first found item with the given attribute `value`."""
        for item in self:
            if getattr(item, name) == value:
                return item
        raise ValueError("Name {!r} not found".format(name))

    def __setattr__(self, name, value):
        """Set value of enumeration item with correct attributes."""
        if isinstance(value, EnumerationItem):
            value = value.__class__(len(self), name, self)
            list.append(self, value)
        return object.__setattr__(self, name, value)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Text formatting actions for :class:`gaupol.Application`."""

import aeidon
import gaupol

class ClearTextsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "clear-texts")
        self.accelerators = ["C"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class CopyTextsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "copy-texts")
        self.accelerators = ["<Control>C"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class CutTextsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "cut-texts")
        self.accelerators = ["<Control>X"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class FindAndReplaceAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "find-and-replace")
        self.accelerators = ["<Control>F", "<Control>H"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)

class FindNextAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "find-next")
        self.accelerators = ["<Control>G"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(application.pattern)

class FindPreviousAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "find-previous")
        self.accelerators = ["<Shift><Control>G"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(application.pattern)

class PasteTextsAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "paste-texts")
        self.accelerators = ["<Control>V"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class ToggleDialogDashesAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "toggle-dialogue-dashes")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class ToggleItalicizationAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "toggle-italicization")
        self.accelerators = ["<Control>I"]
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))
        doc = page.text_column_to_document(col)
        markup = page.project.get_markup(doc)
        aeidon.util.affirm(markup is not None)
        aeidon.util.affirm(markup.italic_tag is not None)

class UseLowerCaseAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "use-lower-case")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class UseSentenceCaseAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "use-sentence-case")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class UseTitleCaseAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "use-title-case")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

class UseUpperCaseAction(gaupol.Action):
    def __init__(self):
        gaupol.Action.__init__(self, "use-upper-case")
        self.action_group = "unsafe"
    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(selected_rows)
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2011 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Mapping position types to basic Python types."""

# aeidon handles frames as integers, seconds as floats and times as strings.
# Some functions accept a "position" argument, which will be interpreted based
# on its type. These functions are introduced here so that callers can use
# 'as_*' to ensure argument types and functions can use 'is_*' to check
# argument types, all in a way which is compatible with the old ambiguous way
# of using int, float and str directly.

__all__ = (
    "as_frame",
    "as_seconds",
    "as_time",
    "is_frame",
    "is_seconds",
    "is_time",
)

def as_frame(pos):
    """Return `pos` as type frame."""
    return int(pos)

def as_seconds(pos):
    """Return `pos` as type seconds."""
    return float(pos)

def as_time(pos):
    """Return `pos` as type time."""
    return str(pos)

def is_frame(pos):
    """Return ``True`` if `pos` is of type frame."""
    return isinstance(pos, int)

def is_seconds(pos):
    """Return ``True`` if `pos` is of type seconds."""
    return isinstance(pos, float)

def is_time(pos):
    """Return ``True`` if `pos` is of type time."""
    return isinstance(pos, str)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Building and updating dynamic menus."""

import aeidon
import gaupol
import os
import sys

from aeidon.i18n   import _
from gi.repository import Gtk


class MenuAgent(aeidon.Delegate):

    """Building and updating dynamic menus."""

    def __init__(self, master):
        """Initialize a :class:`MenuAgent` instance."""
        aeidon.Delegate.__init__(self, master)
        self._columns_popup = None
        self._redo_menu_items = []
        self._tab_popup = None
        self._undo_menu_items = []
        self._view_popup = None
        self._init_signal_handlers()

    @aeidon.deco.once
    def _get_recent_chooser_menu(self):
        """Return a new recent chooser menu."""
        menu = Gtk.RecentChooserMenu()
        menu.set_show_icons(sys.platform != "win32")
        # Can be really slow with network drives.
        # https://github.com/otsaloma/gaupol/issues/175
        menu.set_show_not_found(gaupol.conf.recent.show_not_found)
        menu.set_show_numbers(False)
        menu.set_show_tips(True)
        menu.set_sort_type(Gtk.RecentSortType.MRU)
        menu.connect("item-activated", self._on_recent_menu_item_activated)
        def custom_filter(info, user_data=None):
            application = "gaupol" in info.applications
            mime_type = info.mime_type in [x.mime_type for x in aeidon.formats]
            return bool(application and mime_type)
        recent_filter = Gtk.RecentFilter()
        flags = Gtk.RecentFilterFlags.APPLICATION | Gtk.RecentFilterFlags.MIME_TYPE
        recent_filter.add_custom(flags, custom_filter)
        menu.add_filter(recent_filter)
        menu.set_filter(recent_filter)
        menu.set_limit(10)
        return menu

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        self.connect("page-added", self._update_projects_menu)
        self.connect("page-closed", self._update_projects_menu)
        self.connect("page-saved", self._update_projects_menu)
        self.connect("page-switched", self._update_projects_menu)
        self.connect("pages-reordered", self._update_projects_menu)
        self.connect("init-done", self._update_recent_menus)
        self.connect("page-added", self._update_recent_menus)
        self.connect("page-saved", self._update_recent_menus)

    @aeidon.deco.export
    def _on_activate_project_activate(self, action, parameter):
        """Activate the requested page."""
        index = int(parameter.get_string())
        self.set_current_page(self.pages[index])

    @aeidon.deco.export
    def _on_open_button_show_menu(self, *args):
        """Show a menu listing recent files to open."""
        menu = self._get_recent_chooser_menu()
        self.open_button.set_menu(menu)

    @aeidon.deco.export
    def _on_open_recent_main_file_activate(self, action, *args):
        """Open recent file as main document."""
        if not os.path.isfile(action.gaupol_path):
            return self.flash_message(_("File not found"))
        self.open_main(action.gaupol_path)

    def _on_recent_menu_item_activated(self, chooser, *args):
        """Open recent file as main document."""
        uri = chooser.get_current_uri()
        path = aeidon.util.uri_to_path(uri)
        self.open_button.get_menu().deactivate()
        if not os.path.isfile(path):
            return self.flash_message(_("File not found"))
        self.open_main(path)

    @aeidon.deco.export
    def _on_open_recent_translation_file_activate(self, action, *args):
        """Open recent file as translation document."""
        if not os.path.isfile(action.gaupol_path):
            return self.flash_message(_("File not found"))
        self.open_translation(action.gaupol_path)

    @aeidon.deco.export
    def _on_redo_button_show_menu(self, *args):
        """Show a menu listing all redoable actions."""
        if not self.redo_button.get_menu():
            self.redo_button.set_menu(Gtk.Menu())
        menu = self.redo_button.get_menu()
        for item in menu.get_children():
            menu.remove(item)
        self._redo_menu_items = []
        redoables = []
        with aeidon.util.silent(AttributeError):
            # XXX: Gtk.Actionable.set_action_name doesn't affect the dropdown
            # arrow making it clickable even when there's nothing to redo.
            page = self.get_current_page()
            redoables = page.project.redoables
        for i, action in enumerate(redoables):
            item = Gtk.MenuItem(label=action.description)
            item.gaupol_index = i
            item.connect("activate", self._on_redo_menu_item_activate)
            item.connect("enter-notify-event", self._on_redo_menu_item_enter_notify_event)
            item.connect("leave-notify-event", self._on_redo_menu_item_leave_notify_event)
            self._redo_menu_items.append(item)
            menu.append(item)
        menu.show_all()
        self.redo_button.set_menu(menu)

    def _on_redo_menu_item_activate(self, menu_item):
        """Redo the selected action and all those above it."""
        self.redo(menu_item.gaupol_index + 1)
        self.redo_button.get_menu().deactivate()

    def _on_redo_menu_item_enter_notify_event(self, menu_item, event):
        """Select all actions above `menu_item`."""
        index = menu_item.gaupol_index
        for item in self._redo_menu_items[:index]:
            item.set_state(Gtk.StateType.PRELIGHT)

    def _on_redo_menu_item_leave_notify_event(self, menu_item, event):
        """Unselect all actions above `menu_item`."""
        index = menu_item.gaupol_index
        for item in self._redo_menu_items[:index]:
            item.set_state(Gtk.StateType.NORMAL)

    @aeidon.deco.export
    def _on_tab_widget_button_press_event(self, button, event, page):
        """Display a pop-up menu with tab-related actions."""
        if event.button != 3: return
        if self._tab_popup is None:
            path = os.path.join(aeidon.DATA_DIR, "ui", "tab-popup.ui")
            builder = Gtk.Builder.new_from_file(path)
            self._tab_popup = builder.get_object("tab-popup")
        menu = Gtk.Menu.new_from_model(self._tab_popup)
        menu.attach_to_widget(self.notebook, None)
        menu.popup(parent_menu_shell=None,
                   parent_menu_item=None,
                   func=None,
                   data=None,
                   button=event.button,
                   activate_time=event.time)

        return True

    @aeidon.deco.export
    def _on_undo_button_show_menu(self, *args):
        """Show a menu listing all undoable actions."""
        if not self.undo_button.get_menu():
            self.undo_button.set_menu(Gtk.Menu())
        menu = self.undo_button.get_menu()
        for item in menu.get_children():
            menu.remove(item)
        self._undo_menu_items = []
        undoables = []
        with aeidon.util.silent(AttributeError):
            # XXX: Gtk.Actionable.set_action_name doesn't affect the dropdown
            # arrow making it clickable even when there's nothing to undo.
            page = self.get_current_page()
            undoables = page.project.undoables
        for i, action in enumerate(undoables):
            item = Gtk.MenuItem(label=action.description)
            item.gaupol_index = i
            item.connect("activate", self._on_undo_menu_item_activate)
            item.connect("enter-notify-event", self._on_undo_menu_item_enter_notify_event)
            item.connect("leave-notify-event", self._on_undo_menu_item_leave_notify_event)
            self._undo_menu_items.append(item)
            menu.append(item)
        menu.show_all()
        self.undo_button.set_menu(menu)

    def _on_undo_menu_item_activate(self, menu_item):
        """Undo the selected action and all those above it."""
        self.undo(menu_item.gaupol_index + 1)
        self.undo_button.get_menu().deactivate()

    def _on_undo_menu_item_enter_notify_event(self, menu_item, event):
        """Select all actions above `menu_item`."""
        index = menu_item.gaupol_index
        for item in self._undo_menu_items[:index]:
            item.set_state(Gtk.StateType.PRELIGHT)

    def _on_undo_menu_item_leave_notify_event(self, menu_item, event):
        """Unselect all actions above `menu_item`."""
        index = menu_item.gaupol_index
        for item in self._undo_menu_items[:index]:
            item.set_state(Gtk.StateType.NORMAL)

    @aeidon.deco.export
    def _on_view_button_press_event(self, view, event):
        """Display a right-click pop-up menu to edit data."""
        if event.button != 3: return
        x = int(event.x)
        y = int(event.y)
        value = view.get_path_at_pos(x, y)
        if value is None: return
        path, column, x, y = value
        row = gaupol.util.tree_path_to_row(path)
        if not row in view.get_selected_rows():
            view.set_cursor(path, column)
            view.update_headers()
        if self._view_popup is None:
            path = os.path.join(aeidon.DATA_DIR, "ui", "view-popup.ui")
            builder = Gtk.Builder.new_from_file(path)
            self._view_popup = builder.get_object("view-popup")
        menu = Gtk.Menu.new_from_model(self._view_popup)
        menu.attach_to_widget(view, None)
        menu.popup(parent_menu_shell=None,
                   parent_menu_item=None,
                   func=None,
                   data=None,
                   button=event.button,
                   activate_time=event.time)

        return True

    @aeidon.deco.export
    def _on_view_header_button_press_event(self, button, event):
        """Display a column visibility pop-up menu."""
        if event.button != 3: return
        if self._columns_popup is None:
            path = os.path.join(aeidon.DATA_DIR, "ui", "columns-popup.ui")
            builder = Gtk.Builder.new_from_file(path)
            self._columns_popup = builder.get_object("columns-popup")
        menu = Gtk.Menu.new_from_model(self._columns_popup)
        menu.attach_to_widget(self.get_current_page().view, None)
        menu.popup(parent_menu_shell=None,
                   parent_menu_item=None,
                   func=None,
                   data=None,
                   button=event.button,
                   activate_time=event.time)

        return True

    def _update_projects_menu(self, *args):
        """Update the project menu list of projects."""
        menu = self.get_menubar_section("projects-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        current = self.get_current_page()
        for i, page in enumerate(self.pages):
            label = page.get_main_basename()
            if len(label) > 100:
                label = label[:100] + "…"
            action = "win.activate-project::{:d}".format(i)
            menu.append(label, action)
            if page is current:
                action = self.get_action("activate-project")
                action.set_state(str(i))

    def _update_recent_main_menu(self, *args):
        """Update the file menu list of recent main files."""
        menu = self.get_menubar_section("open-recent-main-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        recent = self._get_recent_chooser_menu()
        for i, uri in enumerate(recent.get_uris()):
            path = aeidon.util.uri_to_path(uri)
            label = os.path.basename(path)
            if len(label) > 100:
                label = label[:100] + "…"
            action = "win.open-recent-main-file-{:d}".format(i)
            menu.append(label, action)
            action = action.replace("win.", "")
            if self.get_action(action):
                # If action i exists, update the file path.
                self.get_action(action).gaupol_path = path
            else:
                # Otherwise, create the action and add to self.window.
                # XXX: For some reason, this can sometimes fail.
                with aeidon.util.silent(Exception):
                    ao = gaupol.Action(action)
                    ao.gaupol_path = path
                    callback = self._on_open_recent_main_file_activate
                    ao.connect("activate", callback)
                    self.window.add_action(ao)

    def _update_recent_menus(self, *args):
        """Update the file menu lists of recent files."""
        self._update_recent_main_menu()
        self._update_recent_translation_menu()
        # Update enabled states of added actions.
        self.update_gui()

    def _update_recent_translation_menu(self, *args):
        """Update the file menu list of recent translation files."""
        menu = self.get_menubar_section("open-recent-translation-placeholder")
        # Menubar not available when running unit tests.
        if menu is None: return
        menu.remove_all()
        recent = self._get_recent_chooser_menu()
        for i, uri in enumerate(recent.get_uris()):
            path = aeidon.util.uri_to_path(uri)
            label = os.path.basename(path)
            if len(label) > 100:
                label = label[:100] + "…"
            action = "win.open-recent-translation-file-{:d}".format(i)
            menu.append(label, action)
            action = action.replace("win.", "")
            if not self.get_action(action):
                ao = gaupol.OpenRecentTranslationFileAction(action)
                callback = self._on_open_recent_translation_file_activate
                ao.connect("activate", callback)
                self.window.add_action(ao)
            self.get_action(action).gaupol_path = path



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""All :mod:`aeidon` error classes."""

__all__ = (
    "Error",
    "AffirmationError",
    "FormatError",
    "ParseError",
    "ProcessError",
)


class Error(Exception):

    """Base class for all :mod:`aeidon` errors."""

    pass


class AffirmationError(Error):

    """
    Something expected to be ``True`` was ``False``.

    :exc:`AffirmationError` is like :exc:`AssertionError`, but without
    the special reliance on :const:`__debug__` and given optimization options.
    :exc:`AffirmationError` can be used to provide essential checks of boolean
    values instead of optional debug checks.
    """

    pass


class FormatError(Error):

    """Unrecognized subtitle file format."""

    pass


class ParseError(Error):

    """
    Failed to parse a subtitle file into positions and texts.

    :exc:`ParseError` is caused by either a syntax error in the file being
    parsed or a programming error in the code doing the parsing.
    """

    pass


class ProcessError(Error):

    """Error in an external process."""

    pass



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer widgets for subtitle data."""

from .float     import FloatCellRenderer
from .integer   import IntegerCellRenderer
from .multiline import MultilineCellRenderer
from .multiline import MultilineDiffCellRenderer
from .time      import TimeCellRenderer

__all__ = (
    "FloatCellRenderer",
    "IntegerCellRenderer",
    "MultilineCellRenderer",
    "MultilineDiffCellRenderer",
    "TimeCellRenderer",
)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tools to edit positions and texts."""

import aeidon
import gaupol

from aeidon.i18n import _
from gi.repository import Gtk


class ToolsAgent(aeidon.Delegate):

    """Tools to edit positions and texts."""

    @aeidon.deco.export
    def _on_adjust_durations_activate(self, *args):
        """Lengthen or shorten durations."""
        dialog = gaupol.DurationAdjustDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_check_spelling_activate(self, *args):
        """Check for incorrect spelling."""
        gaupol.util.set_cursor_busy(self.window)
        try:
            dialog = gaupol.SpellCheckDialog(self.window, self)
        except Exception as error:
            gaupol.util.set_cursor_normal(self.window)
            title = _('Failed to load dictionary for language "{}"')
            title = title.format(gaupol.conf.spell_check.language)
            dialog = gaupol.ErrorDialog(self.window, title, str(error))
            dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
            dialog.set_default_response(Gtk.ResponseType.OK)
            return gaupol.util.flash_dialog(dialog)
        gaupol.util.set_cursor_normal(self.window)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_configure_spell_check_activate(self, *args):
        """Set language and spell-check targets."""
        gaupol.util.set_cursor_busy(self.window)
        dialog = gaupol.LanguageDialog(self.window)
        gaupol.util.set_cursor_normal(self.window)
        gaupol.util.flash_dialog(dialog)
        self.update_gui()

    @aeidon.deco.export
    def _on_convert_framerate_activate(self, *args):
        """Convert framerate."""
        dialog = gaupol.FramerateConvertDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_correct_texts_activate(self, *args):
        """Find and correct errors in texts."""
        gaupol.util.set_cursor_busy(self.window)
        assistant = gaupol.TextAssistant(self.window, self)
        gaupol.util.set_cursor_normal(self.window)
        assistant.show()

    @aeidon.deco.export
    def _on_shift_positions_activate(self, *args):
        """Make subtitles appear earlier or later."""
        page = self.get_current_page()
        if page.edit_mode == aeidon.modes.TIME:
            dialog = gaupol.TimeShiftDialog(self.window, self)
        if page.edit_mode == aeidon.modes.FRAME:
            dialog = gaupol.FrameShiftDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_transform_positions_activate(self, *args):
        """Change positions by linear two-point correction."""
        page = self.get_current_page()
        if page.edit_mode == aeidon.modes.TIME:
            dialog = gaupol.TimeTransformDialog(self.window, self)
        if page.edit_mode == aeidon.modes.FRAME:
            dialog = gaupol.FrameTransformDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Previewing subtitles with a video player."""

import aeidon
import gaupol
import shutil
import tempfile

from aeidon.i18n   import _
from gi.repository import GLib
from gi.repository import Gtk


class PreviewAgent(aeidon.Delegate):

    """Previewing subtitles with a video player."""

    def _check_process_state(self, page, process, fout, command):
        """Check if `process` has terminated or not."""
        if process.poll() is None: return True
        self._handle_output(process, fout, command)
        return False # to not check again.

    def _handle_output(self, process, fout, command):
        """Handle output of finished `process`."""
        fout.close()
        with open(fout.name, "r", errors="replace") as f:
            output = f.read().splitlines()
        output = "\n".join(output[:100])
        print("\nPREVIEW: {}\n{}".format(command, output))
        aeidon.temp.remove(fout.name)
        if process.returncode == 0: return
        dialog = gaupol.PreviewErrorDialog(self.window, output)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def _on_preview_activate(self, *args):
        """Preview from selected position with a video player."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        row = rows[0] if rows else 0
        position = page.project.subtitles[row].start
        col = page.view.get_focus()[1]
        if col == page.view.columns.TRAN_TEXT:
            doc = aeidon.documents.TRAN
        else: # Any other column previews the main file.
            doc = aeidon.documents.MAIN
        self.preview(page, position, doc)

    @aeidon.deco.export
    def preview(self, page, position, doc, temp=False):
        """Preview from `position` with a video player."""
        if (not gaupol.conf.preview.use_custom_command and
            not shutil.which(gaupol.conf.preview.player.executable)):
            return self._show_player_not_found_error_dialog()
        command = gaupol.util.get_preview_command()
        offset = gaupol.conf.preview.offset
        encoding = "utf_8" if gaupol.conf.preview.force_utf_8 else None
        try:
            process, command, fout = page.project.preview(
                position, doc, command, offset, encoding, temp)
        except aeidon.ProcessError as error:
            return self._show_process_error_dialog(str(error))
        except (IOError, OSError) as error:
            return self._show_io_error_dialog(str(error))
        except UnicodeError:
            return self._show_encoding_error_dialog()
        # GLib.child_watch_add does not appear to work on Windows,
        # so let's watch the process by polling it at regular intervals.
        GLib.timeout_add(1000,
                         self._check_process_state,
                         page, process, fout, command)

    @aeidon.deco.export
    def preview_changes(self, page, row, doc, method, args=None, kwargs=None):
        """Preview changes caused by `method` with a video player."""
        subtitles = [x.copy() for x in page.project.subtitles]
        framerate = page.project.framerate
        blocked = page.project.block_all()
        method(register=None, *(args or ()), **(kwargs or {}))
        position = page.project.subtitles[row].start
        page.project.unblock_all(blocked)
        self.preview(page, position, doc, temp=True)
        page.project.set_framerate(framerate, register=None)
        page.project.subtitles = subtitles

    def _show_encoding_error_dialog(self):
        """Show an error dialog after failing to encode file."""
        title = _('Failed to encode subtitle file to temporary directory "{}"').format(tempfile.gettempdir())
        message = _("Subtitle data could not be encoded to a temporary file for preview with the current character encoding. Please first save the subtitle file with a different character encoding.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_io_error_dialog(self, message):
        """Show an error dialog after failing to write file."""
        title = _('Failed to save subtitle file to temporary directory "{}"').format(tempfile.gettempdir())
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_player_not_found_error_dialog(self):
        """Show an error dialog if video player not found."""
        title = _("Video player not found")
        message = _("Please install one of the supported video players: mpv (recommended), MPlayer or VLC. See the preferences dialog to choose the video player or to customize the command.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)

    def _show_process_error_dialog(self, message):
        """Show an error dialog after failing to launch video player."""
        title = _("Failed to launch video player")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(_("_OK"), Gtk.ResponseType.OK)
        dialog.set_default_response(Gtk.ResponseType.OK)
        gaupol.util.flash_dialog(dialog)



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Extension delegates of of :class:`gaupol.Application`."""

from .close    import CloseAgent # noqa
from .edit     import EditAgent # noqa
from .format   import FormatAgent # noqa
from .help     import HelpAgent # noqa
from .menu     import MenuAgent # noqa
from .open     import OpenAgent # noqa
from .preview  import PreviewAgent # noqa
from .save     import SaveAgent # noqa
from .search   import SearchAgent # noqa
from .tools    import ToolsAgent # noqa
from .update   import UpdateAgent # noqa
from .util     import UtilityAgent # noqa
from .video    import VideoAgent # noqa
from .view     import ViewAgent # noqa

__all__ = tuple(x for x in dir() if x.endswith("Agent"))



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Names and ISO 15924 codes for scripts and conversions between them."""

import aeidon
import json
import os

from aeidon.i18n import d_

_scripts = {}


def _init_scripts():
    """Initialize the dictionary mapping codes to names."""
    # Prefer globally installed, fall back on possibly bundled.
    path = "/usr/share/iso-codes/json/iso_15924.json"
    if os.path.isfile(path):
        return _init_scripts_json(path)
    path = os.path.join(aeidon.DATA_DIR, "iso-codes", "iso_15924.json")
    if os.path.isfile(path):
        return _init_scripts_json(path)

def _init_scripts_json(path):
    """Initialize the dictionary mapping codes to names."""
    with open(path, "r", encoding="utf_8") as f:
        iso = json.load(f)
    for script in iso["15924"]:
        code = script.get("alpha_4", None)
        name = script.get("name", None)
        if not code or not name: continue
        _scripts[code] = name

def code_to_name(code):
    """Convert ISO 15924 `code` to localized script name."""
    if not _scripts:
        _init_scripts()
    with aeidon.util.silent(LookupError):
        return d_("iso_15924", _scripts[code])
    return code

def is_valid(code):
    """Return ``True`` if `code` is a valid ISO 15924 script code."""
    if not _scripts:
        _init_scripts()
    return code in _scripts



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base class for dialogs for selecting subtitle files."""

import aeidon
import gaupol

from aeidon.i18n   import _
from gi.repository import Gtk

__all__ = ("FileDialog",)


class FileDialog:

    """Base class for dialogs for selecting subtitle files."""

    _use_autodetection = False

    def get_encoding(self):
        """Return the selected encoding or ``None``."""
        index = self._encoding_combo.get_active()
        if index < 0: return None
        store = self._encoding_combo.get_model()
        return store[index][0]

    def _init_encoding_combo(self):
        """Initialize the character encoding combo box."""
        store = Gtk.ListStore(str, str)
        self._encoding_combo.set_model(store)
        self._populate_encoding_combo()
        view = self._encoding_combo.get_child()
        path = gaupol.util.tree_row_to_path(0)
        view.set_displayed_row(path)
        renderer = Gtk.CellRendererText()
        self._encoding_combo.pack_start(renderer, expand=True)
        self._encoding_combo.add_attribute(renderer, "text", 1)
        func = gaupol.util.separate_combo
        self._encoding_combo.set_row_separator_func(func, None)

    def _init_filters(self):
        """Initialize file filters."""
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("All files"))
        file_filter.add_pattern("*")
        self.add_filter(file_filter)
        file_filter = Gtk.FileFilter()
        file_filter.set_name(_("All supported files"))
        for format in aeidon.formats:
            pattern = "*."
            for x in format.extension[1:]:
                pattern += "[{}{}]".format(x.upper(), x.lower())
            file_filter.add_pattern(pattern)
        self.add_filter(file_filter)
        self.set_filter(file_filter)
        for format in aeidon.formats:
            extension = format.extension
            pattern = "*."
            for x in extension[1:]:
                pattern += "[{}{}]".format(x.upper(), x.lower())
            format = format.label
            name = _("{format} (*{extension})").format(**locals())
            file_filter = Gtk.FileFilter()
            file_filter.set_name(name)
            file_filter.add_pattern(pattern)
            self.add_filter(file_filter)

    def _on_encoding_combo_changed(self, *args):
        """Show the encoding selection dialog."""
        encoding = self.get_encoding()
        if encoding != "other": return
        dialog = gaupol.MenuEncodingDialog(self)
        response = gaupol.util.run_dialog(dialog)
        encoding = dialog.get_encoding()
        visible = dialog.get_visible_encodings()
        dialog.destroy()
        self._encoding_combo.set_active(0)
        if response != Gtk.ResponseType.OK: return
        gaupol.conf.encoding.visible = visible
        if encoding is None: return
        self._populate_encoding_combo(encoding)
        self.set_encoding(encoding)

    def _populate_encoding_combo(self, custom=None):
        """Populate the encoding combo box, including custom encoding."""
        encodings = list(gaupol.conf.encoding.visible)
        locale = aeidon.encodings.get_locale_code()
        encodings.insert(0, locale)
        encodings.append(custom)
        while None in encodings:
            encodings.remove(None)
        encodings = aeidon.util.get_unique(encodings)
        encodings = encodings or ["utf_8"]
        for i, encoding in enumerate(encodings):
            name = aeidon.encodings.code_to_long_name(encoding)
            encodings[i] = (encoding, name)
        if locale is not None:
            name = aeidon.encodings.get_locale_long_name()
            encodings[0] = (locale, name)
        a = 0 if locale is None else 1
        encodings[a:] = sorted(encodings[a:], key=lambda x: x[1])
        separator = gaupol.COMBO_SEPARATOR
        if self._use_autodetection:
            encodings.append((separator, separator))
            encodings.append(("auto", _("Auto-detected")))
        encodings.append((separator, separator))
        encodings.append(("other", _("Other…")))
        self._encoding_combo.get_model().clear()
        store = self._encoding_combo.get_model()
        for encoding in encodings:
            store.append(tuple(encoding))

    def set_encoding(self, encoding):
        """Set the selected encoding."""
        if encoding is None: return
        store = self._encoding_combo.get_model()
        for i in range(len(store)):
            if store[i][0] == encoding:
                return self._encoding_combo.set_active(i)
        if aeidon.encodings.is_valid_code(encoding):
            # Add encoding if not found in store.
            self._populate_encoding_combo(encoding)
            return self.set_encoding(encoding)
        self._encoding_combo.set_active(0)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Changing the visual appearance of application and its documents."""

import aeidon
import gaupol


class ViewAgent(aeidon.Delegate):

    """Changing the visual appearance of application and its documents."""

    @aeidon.deco.export
    def _on_set_edit_mode_activate(self, action, parameter):
        """Change the units in which positions are shown."""
        page = self.get_current_page()
        edit_mode = getattr(aeidon.modes, parameter.get_string())
        if edit_mode == page.edit_mode: return
        gaupol.util.set_cursor_busy(self.window)
        page.edit_mode = edit_mode
        gaupol.conf.editor.mode = edit_mode
        has_focus = page.view.has_focus()
        focus_row, focus_col = page.view.get_focus()
        selected_rows = page.view.get_selected_rows()
        scroller = page.view.get_parent()
        scroller.remove(page.view)
        page.view = gaupol.View(edit_mode)
        self.connect_view_signals(page.view)
        scroller.add(page.view)
        scroller.show_all()
        page.reload_view_all()
        if focus_row is not None:
            page.view.set_focus(focus_row, focus_col)
            page.view.scroll_to_row(focus_row)
        page.view.select_rows(selected_rows)
        page.view.props.has_focus = has_focus
        self.update_gui()
        gaupol.util.set_cursor_normal(self.window)
        page.emit("view-created", page.view)

    @aeidon.deco.export
    def _on_set_framerate_activate(self, action, parameter):
        """Change the framerate to use."""
        page = self.get_current_page()
        framerate = getattr(aeidon.framerates, parameter.get_string())
        if framerate == page.project.framerate: return
        gaupol.util.set_cursor_busy(self.window)
        page.project.set_framerate(framerate, register=None)
        gaupol.conf.editor.framerate = framerate
        if page.edit_mode != page.project.main_file.mode:
            rows = list(range(len(page.project.subtitles)))
            fields = [x for x in gaupol.fields if x.is_position]
            page.reload_view(rows, fields)
        self.update_gui()
        gaupol.util.set_cursor_normal(self.window)

    @aeidon.deco.export
    def _on_set_layout_activate(self, action, parameter):
        """Change the application window layout direction to use."""
        orientation = getattr(gaupol.orientation, parameter.get_string())
        if orientation == gaupol.orientation.HORIZONTAL:
            self.paned.set_orientation(gaupol.orientation.HORIZONTAL)
            self.player_box.orientation = gaupol.orientation.VERTICAL
        if orientation == gaupol.orientation.VERTICAL:
            self.paned.set_orientation(gaupol.orientation.VERTICAL)
            self.player_box.orientation = gaupol.orientation.HORIZONTAL
        gaupol.conf.application_window.layout = orientation
        self.update_gui()

    @aeidon.deco.export
    def _on_toggle_duration_column_activate(self, *args):
        """Show or hide the duration column."""
        self._toggle_column(gaupol.fields.DURATION)

    @aeidon.deco.export
    def _on_toggle_end_column_activate(self, *args):
        """Show or hide the end column."""
        self._toggle_column(gaupol.fields.END)

    @aeidon.deco.export
    def _on_toggle_main_text_column_activate(self, *args):
        """Show or hide the main text column."""
        self._toggle_column(gaupol.fields.MAIN_TEXT)

    @aeidon.deco.export
    def _on_toggle_main_toolbar_activate(self, *args):
        """Show or hide the main toolbar."""
        visible = self.main_toolbar.get_visible()
        self.main_toolbar.set_visible(not visible)
        self.notebook_separator.set_visible(not visible)
        gaupol.conf.application_window.show_main_toolbar = not visible
        self.get_action("toggle-main-toolbar").set_state(not visible)

    @aeidon.deco.export
    def _on_toggle_number_column_activate(self, *args):
        """Show or hide the number column."""
        self._toggle_column(gaupol.fields.NUMBER)

    @aeidon.deco.export
    def _on_toggle_player_activate(self, *args):
        """Show or hide the video player."""
        visible = self.player_box.get_visible()
        self.player_box.set_visible(not visible)
        self.get_action("toggle-player").set_state(not visible)
        self.update_gui()

    @aeidon.deco.export
    def _on_toggle_start_column_activate(self, *args):
        """Show or hide the start column."""
        self._toggle_column(gaupol.fields.START)

    @aeidon.deco.export
    def _on_toggle_translation_text_column_activate(self, *args):
        """Show or hide the translation text column."""
        self._toggle_column(gaupol.fields.TRAN_TEXT)

    def _toggle_column(self, field):
        """Show or hide column corresponding to `field`."""
        page = self.get_current_page()
        col = getattr(page.view.columns, field.name)
        column = page.view.get_column(col)
        visible = column.get_visible()
        gaupol.util.set_cursor_busy(self.window)
        column.set_visible(not visible)
        visible_fields = []
        for field in gaupol.fields:
            col = getattr(page.view.columns, field.name)
            if page.view.get_column(col).get_visible():
                visible_fields.append(field)
        gaupol.conf.editor.visible_fields = visible_fields
        self.update_gui()
        page.view.columns_autosize()
        gaupol.util.set_cursor_normal(self.window)



# -*- coding: utf-8 -*-

# Copyright (C) 2015 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Baseclasses for for user-activatable actions."""

import aeidon

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

__all__ = (
    "Action",
    "OpenRecentTranslationFileAction",
    "RadioAction",
    "ToggleAction",
)


class Action(Gio.SimpleAction):

    """Baseclass for user-activatable actions."""

    def __init__(self, name):
        """Initialize an :class:`Action` instance."""
        GObject.GObject.__init__(self, name=name)
        self.accelerators = []
        self.action_group = "unsafe"

    def _affirm_doable(self, application, page, selected_rows):
        """Raise :exc:`gaupol.AffirmationError` if action cannot be done."""
        pass

    def update_enabled(self, application, page, selected_rows):
        """Update the "enabled" property to match `application`."""
        try:
            self._affirm_doable(application, page, selected_rows)
            self.set_enabled(True)
        except aeidon.AffirmationError:
            self.set_enabled(False)


class OpenRecentTranslationFileAction(Action):

    """Baseclass for a dynamic menu of recent translation files to open."""

    def _affirm_doable(self, application, page, selected_rows):
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.project.main_file is not None)


class RadioAction(Action):

    """Baseclass for user-activatable radio actions."""

    # Gio's abstraction makes radio action instantiation and
    # management of string state values look really stupid
    # in Python. For proper instantiation, subclasses need to
    # define __new__, call new and assign to __class__.

    def activate(self, parameter):
        """Activate action with `parameter`."""
        return Gio.SimpleAction.activate(self, GLib.Variant("s", parameter))

    @staticmethod
    def new(name, parameter_type=GLib.VariantType.new("s")):
        """Return a new radio action."""
        return Gio.SimpleAction.new_stateful(
            name, parameter_type, GLib.Variant("s", ""))

    def get_state(self):
        """Return the current string state."""
        return Action.get_state(self).get_string()

    def set_state(self, value):
        """Set a new string state."""
        Action.set_state(self, GLib.Variant("s", value))


class ToggleAction(Action):

    """Baseclass for user-activatable toggle actions."""

    # Gio's abstraction makes toggle action instantiation and
    # management of boolean state values look really stupid
    # in Python. For proper instantiation, subclasses need to
    # define __new__, call new and assign to __class__.

    @staticmethod
    def new(name, parameter_type=None):
        """Return a new toggle action."""
        return Gio.SimpleAction.new_stateful(
            name, parameter_type, GLib.Variant("b", False))

    def get_state(self):
        """Return the current boolean state."""
        return Action.get_state(self).get_boolean()

    def set_state(self, value):
        """Set a new boolean state."""
        Action.set_state(self, GLib.Variant("b", value))



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Website URLs.

:var BUG_REPORT_URL: Web page where to submit bug reports
:var DOCUMENTATION_URL: Documentation for users
:var HOMEPAGE_URL: Gaupol web pages
:var REGEX_HELP_URL: Documentation on regular expressions
"""

__all__ = (
    "BUG_REPORT_URL",
    "DOCUMENTATION_URL",
    "HOMEPAGE_URL",
    "REGEX_HELP_URL",
)

BUG_REPORT_URL = "https://github.com/otsaloma/gaupol/issues"
DOCUMENTATION_URL = "https://github.com/otsaloma/gaupol/tree/master/doc"
HOMEPAGE_URL = "https://otsaloma.io/gaupol/"
REGEX_HELP_URL = "https://docs.python.org/3/library/re.html#regular-expression-syntax"



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Names and ISO 3166 codes for countries and conversions between them."""

import aeidon
import json
import os

from aeidon.i18n import d_

_countries = {}


def _init_countries():
    """Initialize the dictionary mapping codes to names."""
    # Prefer globally installed, fall back on possibly bundled.
    path = "/usr/share/iso-codes/json/iso_3166-1.json"
    if os.path.isfile(path):
        return _init_countries_json(path)
    path = os.path.join(aeidon.DATA_DIR, "iso-codes", "iso_3166-1.json")
    if os.path.isfile(path):
        return _init_countries_json(path)

def _init_countries_json(path):
    """Initialize the dictionary mapping codes to names."""
    with open(path, "r", encoding="utf_8") as f:
        iso = json.load(f)
    for country in iso["3166-1"]:
        code = country.get("alpha_2", None)
        name = country.get("name", None)
        if not code or not name: continue
        _countries[code] = name

def code_to_name(code):
    """Convert ISO 3166 `code` to localized country name."""
    if not _countries:
        _init_countries()
    with aeidon.util.silent(LookupError):
        return d_("iso_3166", _countries[code])
    return code

def is_valid(code):
    """Return ``True`` if `code` is a valid ISO 3166 country code."""
    if not _countries:
        _init_countries()
    return code in _countries



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting a video file."""

import sys

from aeidon.i18n   import _
from gi.repository import GObject
from gi.repository import Gtk

__all__ = ("VideoDialog",)


class VideoDialog(Gtk.FileChooserDialog):

    """Dialog for selecting a video file."""

    def __init__(self, parent, title, button_label):
        """Initialize a :class:`VideoDialog` instance."""
        GObject.GObject.__init__(self)
        self._init_dialog(parent, title, button_label)
        self._init_filters()

    def _init_dialog(self, parent, title, button_label):
        """Initialize the dialog."""
        self.add_button(_("_Cancel"), Gtk.ResponseType.CANCEL)
        self.add_button(button_label, Gtk.ResponseType.OK)
        self.set_default_response(Gtk.ResponseType.OK)
        self.set_transient_for(parent)
        self.set_title(title)
        self.set_action(Gtk.FileChooserAction.OPEN)

    def _init_filters(self):
        """Intialize the file filters."""
        all_filter = Gtk.FileFilter()
        all_filter.add_pattern("*")
        all_filter.set_name(_("All files"))
        self.add_filter(all_filter)
        video_filter = Gtk.FileFilter()
        video_filter.add_mime_type("video/*")
        video_filter.set_name(_("Video files"))
        self.add_filter(video_filter)
        # Mimetype detection seems unreliable on Windows.
        self.set_filter(all_filter if sys.platform == "win32"
                        else video_filter)



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Help and information."""

import aeidon
import gaupol


class HelpAgent(aeidon.Delegate):

    """Help and information."""

    @aeidon.deco.export
    def _on_browse_documentation_activate(self, *args):
        """Open web browser to view documentation."""
        gaupol.util.show_uri(gaupol.DOCUMENTATION_URL)

    @aeidon.deco.export
    def _on_report_a_bug_activate(self, *args):
        """Open web browser to submit a bug report."""
        gaupol.util.show_uri(gaupol.BUG_REPORT_URL)

    @aeidon.deco.export
    def _on_view_about_dialog_activate(self, *args):
        """Show information about Gaupol."""
        gaupol.util.flash_dialog(gaupol.AboutDialog(self.window))



# -*- coding: utf-8 -*-

# Copyright (C) 2007 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""User-activatable actions for :class:`gaupol.Application`."""

from gaupol.actions.audio    import * # noqa
from gaupol.actions.edit     import * # noqa
from gaupol.actions.file     import * # noqa
from gaupol.actions.help     import * # noqa
from gaupol.actions.projects import * # noqa
from gaupol.actions.text     import * # noqa
from gaupol.actions.tools    import * # noqa
from gaupol.actions.video    import * # noqa
from gaupol.actions.view     import * # noqa

__all__ = tuple(x for x in dir() if x.endswith("Action"))



# -*- coding: utf-8 -*-

# Copyright (C) 2005 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Changing the appearance of texts."""

import aeidon
import re

from aeidon.i18n import _


class FormatAgent(aeidon.Delegate):

    """Changing the appearance of texts."""

    _re_alphanum = re.compile(r"\w")

    @aeidon.deco.export
    @aeidon.deco.revertable
    def add_dialogue_dashes(self, indices, doc, register=-1):
        """Add dialogue dashes to all lines of texts."""
        new_texts = []
        parser = self.get_parser(doc)
        for index in indices:
            subtitle = self.subtitles[index]
            parser.set_text(subtitle.get_text(doc))
            parser.set_regex(r"^[\-\–\—]\s*")
            parser.replacement = ""
            parser.replace_all()
            parser.set_regex(r"^")
            parser.replacement = r"- "
            parser.replace_all()
            new_texts.append(parser.get_text())
        self.replace_texts(indices, doc, new_texts, register=register)
        self.set_action_description(register, _("Adding dialogue dashes"))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def change_case(self, indices, doc, method, register=-1):
        """
        Change the case of texts with `method`.

        `method` should be a string: either "title", "capitalize", "upper" or
        "lower", which correspond to the built-in Python string methods.
        """
        new_texts = []
        parser = self.get_parser(doc)
        for index in indices:
            subtitle = self.subtitles[index]
            parser.set_text(subtitle.get_text(doc))
            self._change_case_first(parser, method)
            new_texts.append(parser.get_text())
        self.replace_texts(indices, doc, new_texts, register=register)
        self.set_action_description(register, _("Changing case"))

    def _change_case_first(self, parser, method):
        """Change the case of the alphanumeric substring."""
        match = self._re_alphanum.search(parser.text)
        if match is None: return
        a = match.start()
        prefix = parser.text[:a]
        text = getattr(parser.text[a:], method)()
        parser.text = prefix + text

    @aeidon.deco.export
    @aeidon.deco.revertable
    def italicize(self, indices, doc, register=-1):
        """Surround texts with italic markup."""
        new_texts = []
        markup = self.get_markup(doc)
        re_italic_tag = markup.italic_tag
        for index in indices:
            text = self.subtitles[index].get_text(doc)
            text = re_italic_tag.sub("", text)
            text = markup.italicize(text)
            new_texts.append(text)
        self.replace_texts(indices, doc, new_texts, register=register)
        self.set_action_description(register, _("Italicizing"))

    @aeidon.deco.export
    @aeidon.deco.revertable
    def remove_dialogue_dashes(self, indices, doc, register=-1):
        """Remove dialogue dashes from all lines of texts."""
        new_texts = []
        parser = self.get_parser(doc)
        for index in indices:
            subtitle = self.subtitles[index]
            parser.set_text(subtitle.get_text(doc))
            parser.set_regex(r"^[\-\–\—]\s*")
            parser.replacement = ""
            parser.replace_all()
            new_texts.append(parser.get_text())
        self.replace_texts(indices, doc, new_texts, register=register)
        self.set_action_description(register, _("Removing dialogue dashes"))

    def _should_add_dialogue_dashes(self, indices, doc):
        """Return ``True`` if dialogue dashes should be added to texts."""
        re_tag = self.get_markup_tag_regex(doc)
        for index in indices:
            text = self.subtitles[index].get_text(doc)
            for line in text.split("\n"):
                # Strip all tags from line.
                # If leftover doesn't start with "-",
                # dialogue dashes should be added.
                if re_tag is not None:
                    line = re_tag.sub("", line)
                if not line.startswith("-"):
                    return True
        return False

    def _should_italicize(self, indices, doc):
        """Return ``True`` if texts should be italicized."""
        re_tag = self.get_markup_tag_regex(doc)
        markup = self.get_markup(doc)
        re_italic_tag = markup.italic_tag
        for index in indices:
            text = self.subtitles[index].get_text(doc)
            # Remove tags from the start of the text,
            # ending after all tags are removed
            # or when an italic tag is found.
            if re_tag is not None:
                while re_tag.match(text) is not None:
                    if re_italic_tag.match(text) is not None:
                        break
                    text = re_tag.sub("", text, 1)
            # If there is no italic tag at the start of the text,
            # all texts should be italicized.
            if re_italic_tag.match(text) is None:
                return True
        return False

    @aeidon.deco.export
    @aeidon.deco.revertable
    def toggle_dialogue_dashes(self, indices, doc, register=-1):
        """Show or hide dialogue dashes on texts."""
        if self._should_add_dialogue_dashes(indices, doc):
            return self.add_dialogue_dashes(indices, doc, register=register)
        return self.remove_dialogue_dashes(indices, doc, register=register)

    @aeidon.deco.export
    @aeidon.deco.revertable
    def toggle_italicization(self, indices, doc, register=-1):
        """Add or remove italic markup surrounding texts."""
        if self._should_italicize(indices, doc):
            return self.italicize(indices, doc, register=register)
        return self.unitalicize(indices, doc, register=register)

    @aeidon.deco.export
    @aeidon.deco.revertable
    def unitalicize(self, indices, doc, register=-1):
        """Remove any italic markup surrounding texts."""
        new_texts = []
        markup = self.get_markup(doc)
        re_italic_tag = markup.italic_tag
        for index in indices:
            text = self.subtitles[index].get_text(doc)
            text = re_italic_tag.sub("", text)
            new_texts.append(text)
        self.replace_texts(indices, doc, new_texts, register=register)
        self.set_action_description(register, _("Unitalicizing"))



# -*- coding: utf-8 -*-

# Copyright (C) 2008 Osmo Salomaa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Finding activating, storing and deactivating extensions."""

import aeidon
import gaupol
import importlib.machinery
import inspect
import os
import re

__all__ = ("ExtensionManager", "ExtensionMetadata",)


class ExtensionManager:

    """
    Finding activating, storing and deactivating extensions.

    :ivar _active: Dictionary mapping module names to extensions
    :ivar _dependants: Dictionary mapping module names to a list of others
    :ivar _metadata: Dictionary mapping module names to metadata items
    :ivar _slaves: List of modules activated to satisfy a dependency
    """
    _global_dir = os.path.join(aeidon.DATA_DIR, "extensions")
    _local_dir  = os.path.join(aeidon.DATA_HOME_DIR, "extensions")
    _re_comment = re.compile(r"#.*$")

    def __init__(self, application):
        """Initialize an :class:`ExtensionManager` instance."""
        self._active = {}
        self.application = application
        self._dependants = {}
        self._metadata = {}
        self._slaves = []

    def find_extensions(self):
        """Find all extensions and parse their metadata files."""
        self._find_extensions_in_directory(self._global_dir)
        self._find_extensions_in_directory(self._local_dir)

    def _find_extensions_in_directory(self, directory):
        """Find all extensions in `directory` and parse their metadata."""
        def is_metadata_file(path):
            return (path.endswith(".extension") or
                    path.endswith(".extension.in"))
        for root, dirs, files in os.walk(directory):
            files = list(filter(is_metadata_file, files))
            for name in [x for x in files if x.endswith(".in")]:
                # If both untranslated and translated metadata files are found,
                # load extension only from the translated one.
                if name[:-3] in files:
                    files.remove(name)
            for name in files:
                path = os.path.abspath(os.path.join(root, name))
                self._parse_metadata(path)

    def get_metadata(self, module):
        """Return an :class:`ExtensionMetadata` instance for `module`."""
        return self._metadata[module]

    def get_modules(self):
        """Return a list of all extension module names."""
        return list(self._metadata.keys())

    def has_help(self, module):
        """Return ``True`` if extension can display documentation."""
        extension = self._active[module]
        return (extension.__class__.show_help is not
                gaupol.Extension.show_help)

    def has_preferences_dialog(self, module):
        """Return True if extension can display a preferences dialog."""
        extension = self._active[module]
        return (extension.__class__.show_preferences_dialog is not
                gaupol.Extension.show_preferences_dialog)

    def is_active(self, module):
        """Return ``True`` if extension is active, ``False`` if not."""
        return (module in self._active)

    def _parse_metadata(self, path):
        """Parse extension metadata file at `path`."""
        lines = aeidon.util.readlines(path, "utf_8", fallback=None)
        lines = [self._re_comment.sub("", x) for x in lines]
        lines = [x.strip() for x in lines]
        metadata = ExtensionMetadata()
        metadata.path = path
        for line in (x for x in lines if x):
            if line.startswith("["): continue
            name, value = line.split("=", 1)
            name = name[1:] if name.startswith("_") else name
            metadata.set_field(name, value)
        module = metadata.get_field("Module")
        self._metadata[module] = metadata

    @aeidon.deco.silent(ImportError, tb=True)
    def setup_extension(self, module, slave=False):
        """Import and setup extension by module name."""
        if module in self._active: return
        metadata = self._metadata[module]
        for dependency in metadata.get_field_list("Requires", ()):
            if not dependency in self._active:
                self.setup_extension(dependency, slave=True)
                self._slaves.append(dependency)
            self._dependants[dependency].append(module)
        directory = os.path.dirname(metadata.path)
        name = "gaupol.extension.{:s}".format(module)
        path = os.path.join(directory, "{}.py".format(module))
        loader = importlib.machinery.SourceFileLoader(name, path)
        module_object = loader.load_module(name)
        for attribute in dir(module_object):
            if attribute.startswith("_"): continue
            value = getattr(module_object, attribute)
            if not inspect.isclass(value): continue
            if issubclass(value, gaupol.Extension):
                extension = value()
                extension.setup(self.application)
                self._active[module] = extension
                self._dependants[module] = []
        self.application.update_gui()

    def setup_extensions(self):
        """Import and setup all extensions configured as active."""
        active = gaupol.conf.extensions.active[:]
        gaupol.conf.extensions.active.clear()
        for module in sorted(set(active) & self._metadata.keys()):
            self.setup_extension(module)
            gaupol.conf.extensions.active.append(module)

    def show_help(self, module):
        """Show documentation on using extension."""
        extension = self._active[module]
        extension.show_help()

    def show_preferences_dialog(self, module, parent):
        """Show a preferences dialog for configuring extension."""
        extension = self._active[module]
        extension.show_preferences_dialog(parent)

    def teardown_extension(self, module):
        """Teardown extension by module name."""
        if not module in self._active: return
        for user in self._dependants[module]:
            # Teardown all extensions depending on this one.
            self.teardown_extension(user)
        if not module in self._active: return
        extension = self._active[module]
        extension.teardown(self.application)
        del self._active[module]
        del self._dependants[module]
        for dependency in list(self._dependants.keys()):
            # Teardown no longer used dependencies.
            if module in self._dependants[dependency]:
                self._dependants[dependency].remove(module)
            if self._dependants[dependency]: continue
            if not dependency in self._slaves: continue
            self.teardown_extension(dependency)
        if module in self._slaves:
            self._slaves.remove(module)
        gaupol.conf.write_to_file()

    def teardown_extensions(self):
        """Teardown all active extensions."""
        for module in list(self._active.keys()):
            self.teardown_extension(module)

    def update_extensions(self, page):
        """Call ``update`` on all active extensions."""
        for name, extension in self._active.items():
            extension.update(self.application, page)


class ExtensionMetadata(aeidon.MetadataItem):

    """Extension metadata specified in a separate desktop-style file."""

    def __init__(self, fields=None):
        """Initialize an :class:`ExtensionMetadata` instance."""
        aeidon.MetadataItem.__init__(self, fields)
        self.path = None
